# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import inspect
import logging
import re
import sys
import urllib.parse
from collections import defaultdict
from typing import Callable, Dict, List, NamedTuple, Optional, Type, Union

from tornado import httpclient, web
from tornado.simple_httpclient import HTTPRequest, HTTPTimeoutError

from ...lib.aio import alru_cache
from ...utils import serialize_serializable, deserialize_serializable

if sys.version_info[:2] == (3, 6):
    # make sure typing works
    re.Pattern = type(re.compile(r".*"))

logger = logging.getLogger(__name__)
_ROOT_PLACEHOLDER = "ROOT_PLACEHOLDER"


class MarsRequestHandler(web.RequestHandler):  # pragma: no cover
    def initialize(self, supervisor_addr: str = None):
        self._supervisor_addr = supervisor_addr


class _WebApiDef(NamedTuple):
    sub_pattern: str
    sub_pattern_compiled: re.Pattern
    method: str
    arg_filter: Optional[Dict] = None


def web_api(
    sub_pattern: str,
    method: Union[str, List[str]],
    arg_filter: Optional[Dict] = None,
    cache_blocking: bool = False,
):
    if not sub_pattern.endswith("$"):  # pragma: no branch
        sub_pattern += "$"
    methods = method if isinstance(method, list) else [method]

    def wrapper(func):
        @functools.wraps(func)
        async def wrapped(self: "MarsServiceWebAPIHandler", *args, **kwargs):
            try:
                if not inspect.iscoroutinefunction(func):
                    return func(self, *args, **kwargs)
                elif not cache_blocking or self.request.method.lower() != "get":
                    res = await func(self, *args, **kwargs)
                else:
                    res = await self._create_or_get_url_future(
                        func, self, *args, **kwargs
                    )
                return res
            except GeneratorExit:
                raise
            except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                exc_type, exc, tb = sys.exc_info()
                err_msg = (
                    f"{exc_type.__name__} when handling request with "
                    f"{type(self).__name__}.{func.__name__}"
                )
                logger.exception(err_msg)
                self.write(serialize_serializable((exc, tb)))
                self.set_status(500, err_msg)

        wrapped._web_api_defs = [
            _WebApiDef(sub_pattern, re.compile(sub_pattern), m, arg_filter)
            for m in methods
        ]
        return wrapped

    return wrapper


@alru_cache(cache_exceptions=False)
async def _get_cluster_api(address: str):
    from ..cluster import ClusterAPI

    return await ClusterAPI.create(address)


@alru_cache(cache_exceptions=False)
async def _get_api_by_key(
    api_cls: Type, session_id: str, address: str, with_key_arg: bool = True
):
    cluster_api = await _get_cluster_api(address)
    [address] = await cluster_api.get_supervisors_by_keys([session_id])
    if with_key_arg:
        return await api_cls.create(session_id, address)
    else:
        return await api_cls.create(address)


class MarsServiceWebAPIHandler(MarsRequestHandler):
    _root_pattern: str = None
    _method_to_handlers: Dict[str, Dict[Callable, _WebApiDef]] = None
    _uri_to_futures: Dict[str, asyncio.Task] = None

    def __init__(self, *args, **kwargs):
        self._collect_services()
        super().__init__(*args, **kwargs)

    def _get_api_by_key(
        self, api_cls: Type, session_id: str, with_key_arg: bool = True
    ):
        return _get_api_by_key(
            api_cls,
            session_id,
            address=self._supervisor_addr,
            with_key_arg=with_key_arg,
        )

    def _create_or_get_url_future(self, func, *args, **kw):
        if self._uri_to_futures is None:
            type(self)._uri_to_futures = dict()

        uri = self.request.uri
        if uri in self._uri_to_futures:
            return self._uri_to_futures[uri]

        def _future_remover(_fut):
            self._uri_to_futures.pop(uri, None)

        task = self._uri_to_futures[uri] = asyncio.create_task(func(*args, **kw))
        task.add_done_callback(_future_remover)
        return task

    @classmethod
    def _collect_services(cls):
        if cls._method_to_handlers is not None:
            return

        cls._method_to_handlers = defaultdict(dict)
        for attr in dir(cls):
            handle_func = getattr(cls, attr, None)
            if not hasattr(handle_func, "_web_api_defs"):
                continue
            web_api_defs = getattr(
                handle_func, "_web_api_defs"
            )  # type: List[_WebApiDef]
            for api_def in web_api_defs:
                cls._method_to_handlers[api_def.method.lower()][handle_func] = api_def

    def prepare(self):
        self.set_header("Content-Type", "application/octet-stream")

    @classmethod
    def get_root_pattern(cls):
        return cls._root_pattern + "(?:/(?P<sub_path>.*)$|$)"

    @functools.lru_cache(100)
    def _route_sub_path(self, http_method: str, sub_path: str):
        handlers = self._method_to_handlers[http_method.lower()]
        method, kwargs = None, None
        for handler_method, web_api_def in handlers.items():
            match = web_api_def.sub_pattern_compiled.match(sub_path)
            if match is not None:
                if web_api_def.arg_filter is not None:
                    if not all(
                        self.get_argument(k, None) == v
                        for k, v in web_api_def.arg_filter.items()
                    ):
                        continue
                    method, kwargs = handler_method, dict(match.groupdict())
                elif method is None:
                    # method matched with arg_filter shall not be overwritten
                    method, kwargs = handler_method, dict(match.groupdict())
        if method is not None:
            return method, kwargs
        else:
            raise web.HTTPError(
                404,
                f"{sub_path} does not match any defined APIs "
                f"with method {http_method.upper()}",
            )

    def _make_handle_http_method(http_method: str):
        async def _handle_http_method(self: "MarsServiceWebAPIHandler", **kwargs):
            # make sure results from APIs is not stored
            self.add_header("Cache-Control", "no-store")

            sub_path = kwargs.pop("sub_path", None) or ""
            method, kw = self._route_sub_path(http_method, sub_path)
            kw.update(kwargs)
            res = method(self, **kw)
            if inspect.isawaitable(res):
                await res

        _handle_http_method.__name__ = http_method.lower()
        return _handle_http_method

    get = _make_handle_http_method("get")
    put = _make_handle_http_method("put")
    post = _make_handle_http_method("post")
    patch = _make_handle_http_method("patch")
    delete = _make_handle_http_method("delete")

    del _make_handle_http_method


class MarsWebAPIClientMixin:
    @property
    def _client(self):
        try:
            return self._client_obj
        except AttributeError:
            self._client_obj = httpclient.AsyncHTTPClient()
            return self._client_obj

    @property
    def request_rewriter(self) -> Callable:
        return getattr(self, "_request_rewriter", None)

    @request_rewriter.setter
    def request_rewriter(self, value: Callable):
        self._request_rewriter = value

    async def _request_url(self, method, path, **kwargs):
        self._running_loop = asyncio.get_running_loop()

        if "data" in kwargs:
            kwargs["body"] = kwargs.pop("data")

        if "params" in kwargs:
            params = kwargs.pop("params")
            for k, v in params.items():
                if isinstance(v, (list, tuple, set)):
                    params[k] = ",".join(str(i) for i in v)
            url_params = urllib.parse.urlencode(params)
            path_connector = "?" if "?" not in path else "&"
            path += path_connector + url_params

        try:
            request = HTTPRequest(path, method=method, **kwargs)
            if self.request_rewriter:
                request = self.request_rewriter(request)
            res = await self._client.fetch(request, raise_error=False)
        except HTTPTimeoutError as ex:
            raise TimeoutError(str(ex)) from None

        if res.code < 400:
            return res
        else:
            exc, tb = None, None
            try:
                exc, tb = deserialize_serializable(res.body)
            except:  # noqa: E722  # nosec  # pylint: disable=bare-except
                pass

            if exc is None:
                raise res.error
            else:
                raise exc.with_traceback(tb)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ..utils import make_import_error_func
from .core import xgboost, XGBScikitLearnBase


XGBRegressor = make_import_error_func("xgboost")
if xgboost:
    from .core import wrap_evaluation_matrices
    from .train import train
    from .predict import predict

    class XGBRegressor(XGBScikitLearnBase):
        """
        Implementation of the scikit-learn API for XGBoost regressor.
        """

        def fit(
            self,
            X,
            y,
            sample_weight=None,
            base_margin=None,
            eval_set=None,
            sample_weight_eval_set=None,
            base_margin_eval_set=None,
            **kw,
        ):
            session = kw.pop("session", None)
            run_kwargs = kw.pop("run_kwargs", dict())
            if kw:
                raise TypeError(
                    f"fit got an unexpected keyword argument '{next(iter(kw))}'"
                )

            dtrain, evals = wrap_evaluation_matrices(
                None,
                X,
                y,
                sample_weight,
                base_margin,
                eval_set,
                sample_weight_eval_set,
                base_margin_eval_set,
            )
            params = self.get_xgb_params()
            self.evals_result_ = dict()
            result = train(
                params,
                dtrain,
                num_boost_round=self.get_num_boosting_rounds(),
                evals=evals,
                evals_result=self.evals_result_,
                session=session,
                run_kwargs=run_kwargs,
            )
            self._Booster = result
            return self

        def predict(self, data, **kw):
            session = kw.pop("session", None)
            run_kwargs = kw.pop("run_kwargs", None)
            return predict(
                self.get_booster(), data, session=session, run_kwargs=run_kwargs, **kw
            )



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from itertools import combinations

from ... import tensor as mt
from ..utils import check_array, check_consistent_length
from ..utils.multiclass import type_of_target


def _average_binary_score(
    binary_metric,
    y_true,
    y_score,
    average,
    sample_weight=None,
    session=None,
    run_kwargs=None,
):
    average_options = (None, "micro", "macro", "weighted", "samples")
    if average not in average_options:  # pragma: no cover
        raise ValueError("average has to be one of {0}".format(average_options))

    y_type = type_of_target(y_true).to_numpy(session=session, **(run_kwargs or dict()))
    if y_type not in ("binary", "multilabel-indicator"):  # pragma: no cover
        raise ValueError("{0} format is not supported".format(y_type))

    if y_type == "binary":
        return binary_metric(y_true, y_score, sample_weight=sample_weight)

    check_consistent_length(
        y_true, y_score, sample_weight, session=session, run_kwargs=run_kwargs
    )
    y_true = check_array(y_true)
    y_score = check_array(y_score)

    not_average_axis = 1
    score_weight = sample_weight
    average_weight = None

    if average == "micro":
        if score_weight is not None:  # pragma: no cover
            score_weight = mt.repeat(score_weight, y_true.shape[1])
        y_true = y_true.ravel()
        y_score = y_score.ravel()

    elif average == "weighted":
        if score_weight is not None:  # pragma: no cover
            average_weight = mt.sum(
                mt.multiply(y_true, mt.reshape(score_weight, (-1, 1))), axis=0
            )
        else:
            average_weight = mt.sum(y_true, axis=0)
        if mt.isclose(average_weight.sum(), 0.0).to_numpy(
            session=session, **(run_kwargs or dict())
        ):
            return 0

    elif average == "samples":
        # swap average_weight <-> score_weight
        average_weight = score_weight
        score_weight = None
        not_average_axis = 0

    if y_true.ndim == 1:
        y_true = y_true.reshape((-1, 1))

    if y_score.ndim == 1:
        y_score = y_score.reshape((-1, 1))

    n_classes = y_score.shape[not_average_axis]
    score = mt.zeros((n_classes,))
    for c in range(n_classes):
        y_true_c = y_true.take([c], axis=not_average_axis).ravel()
        y_score_c = y_score.take([c], axis=not_average_axis).ravel()
        score[c] = binary_metric(y_true_c, y_score_c, sample_weight=score_weight)

    # Average the results
    if average is not None:
        if average_weight is not None:
            # Scores with 0 weights are forced to be 0, preventing the average
            # score from being affected by 0-weighted NaN elements.
            average_weight = mt.asarray(average_weight)
            score[average_weight == 0] = 0
        return mt.average(score, weights=average_weight)
    else:
        return score


def _average_multiclass_ovo_score(
    binary_metric, y_true, y_score, average="macro", session=None, run_kwargs=None
):
    check_consistent_length(y_true, y_score, session=session, run_kwargs=run_kwargs)

    y_true_unique = mt.unique(y_true).to_numpy()
    n_classes = y_true_unique.shape[0]
    n_pairs = n_classes * (n_classes - 1) // 2
    pair_scores = mt.empty(n_pairs)

    is_weighted = average == "weighted"
    prevalence = mt.empty(n_pairs) if is_weighted else None

    # Compute scores treating a as positive class and b as negative class,
    # then b as positive class and a as negative class
    for ix, (a, b) in enumerate(combinations(y_true_unique, 2)):
        a_mask = y_true == a
        b_mask = y_true == b
        ab_mask = mt.logical_or(a_mask, b_mask)

        if is_weighted:
            prevalence[ix] = mt.average(ab_mask)

        a_true = a_mask[ab_mask]
        b_true = b_mask[ab_mask]

        a_true_score = binary_metric(a_true, y_score[ab_mask, a])
        b_true_score = binary_metric(b_true, y_score[ab_mask, b])
        pair_scores[ix] = (a_true_score + b_true_score) / 2

    return mt.average(pair_scores, weights=prevalence)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from dask import is_dask_collection
from dask.array.core import _concatenate2 as array_concat
from dask.dataframe import concat as df_concat
from dask.utils import is_arraylike, is_dataframe_like, is_series_like, is_index_like

from ...remote import spawn


def concat(objs: List):
    """
    Concat the results of partitioned dask task executions. This function guess the
        types of resulting list, then calls the corresponding native dask concat functions.

    Parameters
    ----------
    objs: List
        List of the partitioned dask task execution results, which will be concat.

    Returns
    -------
    obj:
        The concat result

    """
    if is_arraylike(objs[0]):
        res = array_concat(objs, axes=[0])  # TODO: Add concat with args support
    elif any(
        (is_dataframe_like(objs[0]), is_series_like(objs[0]), is_index_like(objs[0]))
    ):
        res = df_concat(objs)
    else:
        res = objs
    return res.compute() if is_dask_collection(res) else res


def reduce(objs: List[List]):
    """
    Spawn a concat task for 2d-list objects

    Parameters
    ----------
    objs: List
        2d-list of the partitioned dask task execution results, which will be concat.

    Returns
    -------
    obj:
        The spawning concat task.
    """
    return spawn(concat, args=([spawn(concat, args=(objs_d,)) for objs_d in objs],))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import scipy.special as spspecial

from ..arithmetic.utils import arithmetic_operand
from ..utils import infer_dtype, implement_scipy
from .core import (
    TensorSpecialUnaryOp,
    TensorSpecialMultiOp,
    TensorSpecialBinOp,
    _register_special_op,
)


class NoOrderSpecialMixin:
    @classmethod
    def _get_func(cls, xp):
        func = super()._get_func(xp)

        def _wrapped(*args, **kw):
            kw.pop("order", None)
            return func(*args, **kw)

        return _wrapped


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorGamma(TensorSpecialUnaryOp):
    _func_name = "gamma"


@implement_scipy(spspecial.gamma)
@infer_dtype(spspecial.gamma)
def gamma(x, **kwargs):
    op = TensorGamma(**kwargs)
    return op(x)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorGammaln(TensorSpecialUnaryOp):
    _func_name = "gammaln"


@implement_scipy(spspecial.gammaln)
@infer_dtype(spspecial.gammaln)
def gammaln(x, out=None, where=None, **kwargs):
    """
    Logarithm of the absolute value of the Gamma function.

    Parameters
    ----------
    x : array-like
        Values on the real line at which to compute ``gammaln``
    out : Tensor, None, or tuple of Tensor and None, optional
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or `None`,
        a freshly-allocated tensor is returned. A tuple (possible only as a
        keyword argument) must have length equal to the number of outputs.
    where : array_like, optional
        Values of True indicate to calculate the ufunc at that position, values
        of False indicate to leave the value in the output alone.
    **kwargs

    Returns
    -------
    gammaln : Tensor
        Values of ``gammaln`` at x.

    See Also
    --------
    gammasgn : sign of the gamma function
    loggamma : principal branch of the logarithm of the gamma function

    Notes
    -----
    When used in conjunction with `gammasgn`, this function is useful
    for working in logspace on the real axis without having to deal with
    complex numbers, via the relation ``exp(gammaln(x)) = gammasgn(x)*gamma(x)``.

    For complex-valued log-gamma, use `loggamma` instead of `gammaln`.
    """
    op = TensorGammaln(**kwargs)
    return op(x, out=out, where=where)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorLogGamma(TensorSpecialUnaryOp):
    _func_name = "loggamma"


@implement_scipy(spspecial.loggamma)
@infer_dtype(spspecial.loggamma)
def loggamma(x, **kwargs):
    op = TensorLogGamma(**kwargs)
    return op(x)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorGammaSgn(TensorSpecialUnaryOp):
    _func_name = "gammasgn"


@implement_scipy(spspecial.gammasgn)
@infer_dtype(spspecial.gammasgn)
def gammasgn(x, **kwargs):
    op = TensorGammaSgn(**kwargs)
    return op(x)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorGammaInc(TensorSpecialBinOp):
    _func_name = "gammainc"


@implement_scipy(spspecial.gammainc)
@infer_dtype(spspecial.gammainc)
def gammainc(a, b, **kwargs):
    op = TensorGammaInc(**kwargs)
    return op(a, b)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorGammaIncInv(TensorSpecialBinOp):
    _func_name = "gammaincinv"


@implement_scipy(spspecial.gammaincinv)
@infer_dtype(spspecial.gammaincinv)
def gammaincinv(a, b, **kwargs):
    op = TensorGammaIncInv(**kwargs)
    return op(a, b)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorGammaIncc(TensorSpecialBinOp):
    _func_name = "gammaincc"


@implement_scipy(spspecial.gammainc)
@infer_dtype(spspecial.gammainc)
def gammaincc(a, b, **kwargs):
    op = TensorGammaIncc(**kwargs)
    return op(a, b)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorGammaInccInv(TensorSpecialBinOp):
    _func_name = "gammainccinv"


@implement_scipy(spspecial.gammainccinv)
@infer_dtype(spspecial.gammainccinv)
def gammainccinv(a, b, **kwargs):
    op = TensorGammaInccInv(**kwargs)
    return op(a, b)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorBeta(TensorSpecialBinOp):
    _func_name = "beta"


@implement_scipy(spspecial.beta)
@infer_dtype(spspecial.beta)
def beta(a, b, out=None, **kwargs):
    op = TensorBeta(**kwargs)
    return op(a, b, out=out)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorBetaLn(TensorSpecialBinOp):
    _func_name = "betaln"


@implement_scipy(spspecial.betaln)
@infer_dtype(spspecial.betaln)
def betaln(a, b, out=None, **kwargs):
    op = TensorBetaLn(**kwargs)
    return op(a, b, out=out)


@_register_special_op
class TensorBetaInc(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "betainc"


@implement_scipy(spspecial.betainc)
@infer_dtype(spspecial.betainc)
def betainc(a, b, x, out=None, **kwargs):
    op = TensorBetaInc(**kwargs)
    return op(a, b, x, out=out)


@_register_special_op
class TensorBetaIncInv(TensorSpecialMultiOp):
    _ARG_COUNT = 3
    _func_name = "betaincinv"


@implement_scipy(spspecial.betaincinv)
@infer_dtype(spspecial.betaincinv)
def betaincinv(a, b, y, out=None, **kwargs):
    op = TensorBetaIncInv(**kwargs)
    return op(a, b, y, out=out)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorPsi(TensorSpecialUnaryOp):
    _func_name = "psi"


@implement_scipy(spspecial.psi)
@infer_dtype(spspecial.psi)
def psi(x, out=None, **kwargs):
    op = TensorPsi(**kwargs)
    return op(x, out=out)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorRGamma(TensorSpecialUnaryOp):
    _func_name = "rgamma"


@implement_scipy(spspecial.rgamma)
@infer_dtype(spspecial.rgamma)
def rgamma(x, out=None, **kwargs):
    op = TensorRGamma(**kwargs)
    return op(x, out=out)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorPolyGamma(NoOrderSpecialMixin, TensorSpecialBinOp):
    _func_name = "polygamma"


@implement_scipy(spspecial.polygamma)
@infer_dtype(spspecial.polygamma)
def polygamma(a, b, **kwargs):
    op = TensorPolyGamma(**kwargs)
    return op(a, b)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorMultiGammaLn(NoOrderSpecialMixin, TensorSpecialBinOp):
    _func_name = "multigammaln"


@implement_scipy(spspecial.multigammaln)
@infer_dtype(spspecial.multigammaln)
def multigammaln(a, b, **kwargs):
    op = TensorMultiGammaLn(**kwargs)
    return op(a, b)


@_register_special_op
@arithmetic_operand(sparse_mode="unary")
class TensorDiGamma(TensorSpecialUnaryOp):
    _func_name = "digamma"


@implement_scipy(spspecial.digamma)
@infer_dtype(spspecial.digamma)
def digamma(x, out=None, **kwargs):
    op = TensorDiGamma(**kwargs)
    return op(x, out=out)


@_register_special_op
@arithmetic_operand(sparse_mode="binary_and")
class TensorPoch(TensorSpecialBinOp):
    _func_name = "poch"


@implement_scipy(spspecial.poch)
@infer_dtype(spspecial.poch)
def poch(a, b, **kwargs):
    op = TensorPoch(**kwargs)
    return op(a, b)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd

from ...config import options
from ...core import recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import (
    FieldTypes,
    AnyField,
    BoolField,
    Int32Field,
    Int64Field,
    ListField,
    StringField,
)
from ...utils import ceildiv
from ..operands import DataFrameOperand
from ..utils import parse_index


class DataFrameSortOperand(DataFrameOperand):
    axis = Int32Field("axis")
    ascending = AnyField("ascending")
    inplace = BoolField("inplace")
    kind = StringField("kind")
    na_position = StringField("na_position")
    ignore_index = BoolField("ignore_index")
    parallel_kind = StringField("parallel_kind")
    psrs_kinds = ListField("psrs_kinds", FieldTypes.string)
    nrows = Int64Field("nrows", default=None)

    @classmethod
    def _tile_head(cls, op: "DataFrameSortOperand"):
        from ..merge import DataFrameConcat

        inp = op.inputs[0]
        out = op.outputs[0]
        axis = op.axis
        assert axis == 0
        pd_index = out.index_value.to_pandas()
        combine_size = options.combine_size

        if inp.ndim == 2:
            if inp.chunk_shape[1 - axis] > 1:  # pragma: no cover
                if any(pd.isna(s) for s in inp.nsplits[1 - axis]):
                    yield
                inp = yield from recursive_tile(
                    inp.rechunk({1 - axis: inp.shape[1 - axis]})
                )

        out_chunks = []
        for c in inp.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op.stage = OperandStage.map
            chunk_params = c.params
            chunk_params["index_value"] = parse_index(pd_index, c)
            out_chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))

        while True:
            chunk_size = ceildiv(len(out_chunks), combine_size)
            combine_chunks = []
            for i in range(chunk_size):
                chunk_index = (i,) if inp.ndim == 1 else (i, 0)

                to_combine_chunks = out_chunks[
                    i * combine_size : (i + 1) * combine_size
                ]
                concat_params = to_combine_chunks[0].params
                concat_params["index"] = chunk_index
                shape = list(to_combine_chunks[0].shape)
                shape[0] = sum(c.shape[0] for c in to_combine_chunks)
                shape = tuple(shape)
                concat_params["shape"] = shape
                c = DataFrameConcat(axis=axis, output_types=op.output_types).new_chunk(
                    to_combine_chunks, kws=[concat_params]
                )

                chunk_op = op.copy().reset_key()
                chunk_op.stage = (
                    OperandStage.combine if chunk_size > 1 else OperandStage.agg
                )
                chunk_params = c.params
                chunk_params["index_value"] = parse_index(pd_index, c)
                chunk_params["shape"] = (min(shape[0], op.nrows),) + shape[1:]
                combine_chunks.append(chunk_op.new_chunk([c], kws=[chunk_params]))
            out_chunks = combine_chunks
            if chunk_size == 1:
                break

        new_op = op.copy()
        params = out.params
        params["nsplits"] = tuple((s,) for s in out.shape)
        params["chunks"] = out_chunks
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _tile(cls, op):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def tile(cls, op: "DataFrameSortOperand"):
        if op.nrows is not None:
            return (yield from cls._tile_head(op))
        else:
            return (yield from cls._tile(op))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import copy
import heapq
import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import DefaultDict, Dict, List, Optional, Tuple, Union, Set

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....metrics import Metrics
from ....resource import ZeroResource
from ....utils import dataslots
from ...subtask import Subtask
from ...task import TaskAPI
from ..utils import redirect_subtask_errors

logger = logging.getLogger(__name__)

_DEFAULT_SUBMIT_PERIOD = 0


@dataslots
@dataclass
class HeapItem:
    subtask: Subtask
    priority: Tuple

    def __lt__(self, other: "HeapItem"):
        return self.priority > other.priority


class SubtaskQueueingActor(mo.Actor):
    _stid_to_bands: DefaultDict[str, List[Tuple]]
    _stid_to_items: Dict[str, HeapItem]
    _band_queues: DefaultDict[Tuple, List[HeapItem]]

    @classmethod
    def gen_uid(cls, session_id: str):
        return f"{session_id}_subtask_queueing"

    def __init__(self, session_id: str, submit_period: Union[float, int] = None):
        self._session_id = session_id
        self._stid_to_bands = defaultdict(list)
        self._stid_to_items = dict()
        # Note that we need to ensure top item in every band heap queue is valid,
        # so that we can ensure band queue is busy if the band queue is not empty.
        self._band_queues = defaultdict(list)

        self._cluster_api = None
        self._slots_ref = None
        self._assigner_ref = None

        self._band_to_resource = dict()
        self._band_watch_task = None
        self._max_enqueue_id = 0

        self._periodical_submit_task = None
        self._submit_period = submit_period or _DEFAULT_SUBMIT_PERIOD
        self._submitted_subtask_number = Metrics.gauge(
            "mars.band.submitted_subtask_number",
            "The number of submitted subtask to a band.",
            ("session_id", "band"),
        )
        self._unsubmitted_subtask_number = Metrics.gauge(
            "mars.band.unsubmitted_subtask_number",
            "The number of unsubmitted subtask to a band.",
            ("session_id", "band"),
        )

    async def __post_create__(self):
        from ...cluster import ClusterAPI

        self._cluster_api = await ClusterAPI.create(self.address)
        self._band_to_resource = {}

        async def watch_bands():
            async for bands in self._cluster_api.watch_all_bands():
                # confirm ready bands indeed changed
                if bands != self._band_to_resource:
                    old_band_resource = self._band_to_resource
                    self._band_to_resource = copy.deepcopy(bands)
                    if self._band_queues:
                        await self.balance_queued_subtasks()
                        # Refresh global slot manager to get latest bands,
                        # so that subtasks reassigned to the new bands can be
                        # ensured to get submitted as least one subtask every band
                        # successfully.
                        await self._slots_ref.refresh_bands()
                        all_bands = {*bands.keys(), *old_band_resource.keys()}
                        bands_delta = {}
                        for b in all_bands:
                            new_resource = bands.get(b, ZeroResource)
                            old_resource = old_band_resource.get(b, ZeroResource)
                            delta = new_resource - old_resource
                            if delta != ZeroResource:
                                bands_delta[b] = delta
                        # Submit tasks on new bands manually, otherwise some subtasks
                        # will never got submitted. Note that we must ensure every new
                        # band will get at least one subtask submitted successfully.
                        # Later subtasks submit on the band will be triggered by the
                        # success of previous subtasks on the same band.
                        logger.info(
                            "Bands changed with delta %s, submit all bands.",
                            bands_delta,
                        )
                        await self.ref().submit_subtasks()

        self._band_watch_task = asyncio.create_task(watch_bands())

        from .globalresource import GlobalResourceManagerActor

        [self._slots_ref] = await self._cluster_api.get_supervisor_refs(
            [GlobalResourceManagerActor.default_uid()]
        )
        from .assigner import AssignerActor

        self._assigner_ref = await mo.actor_ref(
            AssignerActor.gen_uid(self._session_id), address=self.address
        )

        if self._submit_period > 0:
            self._periodical_submit_task = self.ref().periodical_submit.tell_delay(
                delay=self._submit_period
            )

    async def __pre_destroy__(self):
        self._band_watch_task.cancel()
        if self._periodical_submit_task is not None:  # pragma: no branch
            self._periodical_submit_task.cancel()

    async def periodical_submit(self):
        await self.ref().submit_subtasks.tell()
        self._periodical_submit_task = self.ref().periodical_submit.tell_delay(
            delay=self._submit_period
        )

    @alru_cache
    async def _get_task_api(self):
        return await TaskAPI.create(self._session_id, self.address)

    @alru_cache(cache_exceptions=False)
    async def _get_manager_ref(self):
        from .manager import SubtaskManagerActor

        return await mo.actor_ref(
            SubtaskManagerActor.gen_uid(self._session_id), address=self.address
        )

    async def add_subtasks(
        self,
        subtasks: List[Subtask],
        priorities: List[Tuple],
        exclude_bands: Set[Tuple] = None,
        random_when_unavailable: bool = True,
    ):
        bands = await self._assigner_ref.assign_subtasks(
            subtasks, exclude_bands, random_when_unavailable
        )
        for subtask, band, priority in zip(subtasks, bands, priorities):
            assert band is not None
            self._stid_to_bands[subtask.subtask_id].append(band)
            heap_item = self._stid_to_items[subtask.subtask_id] = HeapItem(
                subtask, priority + (self._max_enqueue_id,)
            )
            self._max_enqueue_id += 1
            heapq.heappush(self._band_queues[band], heap_item)
            logger.debug(
                "Subtask %s enqueued to band %s excluded from %s.",
                subtask.subtask_id,
                band,
                exclude_bands,
            )
        logger.debug("%d subtasks enqueued", len(subtasks))

    async def submit_subtasks(self, band: Tuple = None, limit: Optional[int] = None):
        logger.debug("Submitting subtasks with limit %s", limit)

        if not limit and band not in self._band_to_resource:
            self._band_to_resource = await self._cluster_api.get_all_bands()

        bands = [band] if band is not None else list(self._band_to_resource.keys())
        submit_aio_tasks = []
        manager_ref = await self._get_manager_ref()

        apply_delays = []
        submit_items_list = []
        submitted_bands = []

        for band in bands:
            band_limit = limit or (
                self._band_to_resource[band].num_cpus
                or self._band_to_resource[band].num_gpus
            )
            task_queue = self._band_queues[band]
            submit_items = dict()
            while (
                self._ensure_top_item_valid(task_queue)
                and len(submit_items) < band_limit
            ):
                item = heapq.heappop(task_queue)
                submit_items[item.subtask.subtask_id] = item

            subtask_ids = list(submit_items)
            if not subtask_ids:
                continue

            submitted_bands.append(band)
            submit_items_list.append(submit_items)

            # Before hbo, when a manager finish a subtask, it will schedule one subtask successfully because
            # there is a slot idle. But now we have memory requirements, so the subtask may apply resource
            # from supervisor failed. In such cases, those subtasks will never got scheduled.
            # TODO We can use `_periodical_submit_task` to submit those subtasks.
            subtask_resources = [
                item.subtask.required_resource for item in submit_items.values()
            ]
            apply_delays.append(
                self._slots_ref.apply_subtask_resources.delay(
                    band, self._session_id, subtask_ids, subtask_resources
                )
            )

        async with redirect_subtask_errors(
            self,
            [
                item.subtask
                for submit_items in submit_items_list
                for item in submit_items.values()
            ],
        ):
            submitted_ids_list = await self._slots_ref.apply_subtask_resources.batch(
                *apply_delays
            )

        for band, submit_items, submitted_ids in zip(
            submitted_bands, submit_items_list, submitted_ids_list
        ):
            subtask_ids = list(submit_items)
            task_queue = self._band_queues[band]

            async with redirect_subtask_errors(
                self, [item.subtask for item in submit_items.values()]
            ):
                non_submitted_ids = [k for k in submit_items if k not in submitted_ids]
                tags = {
                    "session_id": self._session_id,
                    "band": band[0] if band else "",
                }
                self._submitted_subtask_number.record(len(submitted_ids), tags)
                self._unsubmitted_subtask_number.record(len(non_submitted_ids), tags)
                if submitted_ids:
                    for stid in subtask_ids:
                        if stid not in submitted_ids:
                            continue
                        item = submit_items[stid]
                        logger.debug("Submit subtask %r to band %r", item.subtask, band)
                        submit_aio_tasks.append(
                            asyncio.create_task(
                                manager_ref.submit_subtask_to_band.tell(
                                    item.subtask.subtask_id, band
                                )
                            )
                        )
                        await asyncio.sleep(0)
                        self.remove_queued_subtasks([item.subtask.subtask_id])
                else:
                    logger.debug("No slots available")

            for stid in non_submitted_ids:
                # TODO if subtasks submit failed due to lacking memory/cpu/gpu resources, lower the priority so that
                # other subtasks can be submitted.
                heapq.heappush(task_queue, submit_items[stid])

        if submit_aio_tasks:
            yield asyncio.gather(*submit_aio_tasks)

    def _ensure_top_item_valid(self, task_queue):
        """Clean invalid subtask item from the queue to ensure that when the queue is not empty,
        there is always some subtasks waiting being scheduled."""
        while (
            task_queue and task_queue[0].subtask.subtask_id not in self._stid_to_items
        ):
            #  skip removed items (as they may be re-pushed into the queue)
            heapq.heappop(task_queue)
        return bool(task_queue)

    @mo.extensible
    def update_subtask_priority(self, subtask_id: str, priority: Tuple):
        if subtask_id not in self._stid_to_bands:
            return
        for band in self._stid_to_bands[subtask_id]:
            new_item = HeapItem(self._stid_to_items[subtask_id].subtask, priority)
            self._stid_to_items[subtask_id] = new_item
            heapq.heappush(self._band_queues[band], new_item)

    def remove_queued_subtasks(self, subtask_ids: List[str]):
        for stid in subtask_ids:
            bands = self._stid_to_bands.pop(stid, [])
            self._stid_to_items.pop(stid, None)
            for band in bands:
                band_queue = self._band_queues.get(band)
                self._ensure_top_item_valid(band_queue)

    async def all_bands_busy(self) -> bool:
        """Return True if all bands queue has tasks waiting to be submitted."""
        bands = set(self._band_to_resource.keys())
        if set(self._band_queues.keys()).issuperset(bands):
            return all(len(self._band_queues[band]) > 0 for band in bands)
        return False

    async def balance_queued_subtasks(self):
        # record length of band queues
        band_num_queued_subtasks = {
            band: len(queue) for band, queue in self._band_queues.items()
        }
        move_queued_subtasks = await self._assigner_ref.reassign_subtasks(
            band_num_queued_subtasks
        )
        items = []
        # rewrite band queues according to feedbacks from assigner
        for band, move in move_queued_subtasks.items():
            task_queue = self._band_queues[band]
            assert move + len(task_queue) >= 0
            for _ in range(abs(move)):
                if move < 0:
                    # TODO: pop item of low priority
                    item = heapq.heappop(task_queue)
                    self._stid_to_bands[item.subtask.subtask_id].remove(band)
                    items.append(item)
                elif move > 0:
                    item = items.pop()
                    self._stid_to_bands[item.subtask.subtask_id].append(band)
                    heapq.heappush(task_queue, item)
            if len(task_queue) == 0:
                self._band_queues.pop(band)



# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import functools
import logging
import random
import time
import uuid
from urllib.parse import urlparse

from ...lib.aio import new_isolation, stop_isolation
from ...services.cluster.api import WebClusterAPI
from ...session import new_session
from ...utils import calc_size_by_str
from ..utils import wait_services_ready
from .config import (
    NamespaceConfig,
    RoleConfig,
    RoleBindingConfig,
    ServiceConfig,
    MarsSupervisorsConfig,
    MarsWorkersConfig,
)

try:
    from kubernetes.client.rest import ApiException as K8SApiException
except ImportError:  # pragma: no cover
    K8SApiException = None

logger = logging.getLogger(__name__)


class KubernetesClusterClient:
    def __init__(self, cluster):
        self._cluster = cluster
        self._endpoint = None
        self._session = None

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def namespace(self):
        return self._cluster.namespace

    @property
    def session(self):
        return self._session

    def start(self):
        try:
            self._endpoint = self._cluster.start()
            self._session = new_session(self._endpoint)
        except:  # noqa: E722  # nosec  # pylint: disable=bare-except
            self.stop()
            raise

    def stop(self, wait=False, timeout=0):
        self._cluster.stop(wait=wait, timeout=timeout)


class KubernetesCluster:
    _supervisor_config_cls = MarsSupervisorsConfig
    _worker_config_cls = MarsWorkersConfig
    _default_service_port = 7103
    _default_web_port = 7104

    def __init__(
        self,
        kube_api_client=None,
        image=None,
        namespace=None,
        supervisor_num=1,
        supervisor_cpu=1,
        supervisor_mem="4G",
        supervisor_mem_limit_ratio=None,
        worker_num=1,
        worker_cpu=None,
        worker_mem=None,
        worker_spill_paths=None,
        worker_cache_mem=None,
        min_worker_num=None,
        worker_min_cache_mem=None,
        worker_mem_limit_ratio=None,
        web_port=None,
        service_name=None,
        service_type=None,
        timeout=None,
        **kwargs,
    ):
        from kubernetes import client as kube_client

        if worker_cpu is None or worker_mem is None:  # pragma: no cover
            raise TypeError("`worker_cpu` and `worker_mem` must be specified")

        self._api_client = kube_api_client
        self._core_api = kube_client.CoreV1Api(kube_api_client)

        self._namespace = namespace
        self._image = image
        self._timeout = timeout
        self._service_name = service_name or "marsservice"
        self._service_type = service_type or "NodePort"
        self._extra_volumes = kwargs.pop("extra_volumes", ())
        self._pre_stop_command = kwargs.pop("pre_stop_command", None)
        self._log_when_fail = kwargs.pop("log_when_fail", False)

        extra_modules = kwargs.pop("extra_modules", None) or []
        extra_modules = (
            extra_modules.split(",")
            if isinstance(extra_modules, str)
            else extra_modules
        )
        extra_envs = kwargs.pop("extra_env", None) or dict()
        extra_labels = kwargs.pop("extra_labels", None) or dict()
        service_port = kwargs.pop("service_port", None) or self._default_service_port

        def _override_modules(updates):
            modules = set(extra_modules)
            updates = updates.split(",") if isinstance(updates, str) else updates
            modules.update(updates)
            return sorted(modules)

        def _override_dict(d, updates):
            updates = updates or dict()
            ret = d.copy()
            ret.update(updates)
            return ret

        _override_envs = functools.partial(_override_dict, extra_envs)
        _override_labels = functools.partial(_override_dict, extra_labels)

        self._supervisor_num = supervisor_num
        self._supervisor_cpu = supervisor_cpu
        self._supervisor_mem = calc_size_by_str(supervisor_mem, None)
        self._supervisor_mem_limit_ratio = supervisor_mem_limit_ratio
        self._supervisor_extra_modules = _override_modules(
            kwargs.pop("supervisor_extra_modules", [])
        )
        self._supervisor_extra_env = _override_envs(
            kwargs.pop("supervisor_extra_env", None)
        )
        self._supervisor_extra_labels = _override_labels(
            kwargs.pop("supervisor_extra_labels", None)
        )
        self._supervisor_service_port = (
            kwargs.pop("supervisor_service_port", None) or service_port
        )
        self._web_port = web_port or self._default_web_port
        self._external_web_endpoint = None

        self._worker_num = worker_num
        self._worker_cpu = worker_cpu
        self._worker_mem = calc_size_by_str(worker_mem, None)
        self._worker_mem_limit_ratio = worker_mem_limit_ratio
        self._worker_spill_paths = worker_spill_paths
        self._worker_cache_mem = worker_cache_mem
        self._worker_min_cache_men = worker_min_cache_mem
        self._min_worker_num = min_worker_num
        self._worker_extra_modules = _override_modules(
            kwargs.pop("worker_extra_modules", [])
        )
        self._worker_extra_env = _override_envs(kwargs.pop("worker_extra_env", None))
        self._worker_extra_labels = _override_labels(
            kwargs.pop("worker_extra_labels", None)
        )
        self._worker_service_port = (
            kwargs.pop("worker_service_port", None) or service_port
        )

    @property
    def namespace(self):
        return self._namespace

    def _get_free_namespace(self):
        while True:
            namespace = "mars-ns-" + str(uuid.uuid4().hex)
            try:
                self._core_api.read_namespace(namespace)
            except K8SApiException as ex:
                if ex.status != 404:  # pragma: no cover
                    raise
                return namespace

    def _create_kube_service(self):
        if self._service_type != "NodePort":  # pragma: no cover
            raise NotImplementedError(
                f"Service type {self._service_type} not supported"
            )

        service_config = ServiceConfig(
            self._service_name,
            service_type="NodePort",
            port=self._web_port,
            selector={"mars/service-type": MarsSupervisorsConfig.rc_name},
        )
        self._core_api.create_namespaced_service(
            self._namespace, service_config.build()
        )

    def _get_ready_pod_count(self, label_selector):
        query = self._core_api.list_namespaced_pod(
            namespace=self._namespace, label_selector=label_selector
        ).to_dict()
        cnt = 0
        for el in query["items"]:
            if el["status"]["phase"] in ("Error", "Failed"):
                logger.warning(
                    "Error in starting pod, message: %s", el["status"]["message"]
                )
                continue
            if "status" not in el or "conditions" not in el["status"]:
                cnt += 1
            elif any(
                cond["type"] == "Ready" and cond["status"] == "True"
                for cond in el["status"].get("conditions") or ()
            ):
                cnt += 1
        return cnt

    def _create_namespace(self):
        if self._namespace is None:
            namespace = self._namespace = self._get_free_namespace()
        else:
            namespace = self._namespace

        self._core_api.create_namespace(NamespaceConfig(namespace).build())

    def _create_roles_and_bindings(self):
        # create role and binding
        role_config = RoleConfig(
            "mars-pod-operator",
            self._namespace,
            api_groups="",
            resources="pods,endpoints,services",
            verbs="get,watch,list,patch",
        )
        role_config.create_namespaced(self._api_client, self._namespace)
        role_binding_config = RoleBindingConfig(
            "mars-pod-operator-binding", self._namespace, "mars-pod-operator", "default"
        )
        role_binding_config.create_namespaced(self._api_client, self._namespace)

    def _create_supervisors(self):
        supervisors_config = self._supervisor_config_cls(
            self._supervisor_num,
            image=self._image,
            cpu=self._supervisor_cpu,
            memory=self._supervisor_mem,
            memory_limit_ratio=self._supervisor_mem_limit_ratio,
            modules=self._supervisor_extra_modules,
            volumes=self._extra_volumes,
            service_name=self._service_name,
            service_port=self._supervisor_service_port,
            web_port=self._web_port,
            pre_stop_command=self._pre_stop_command,
        )
        supervisors_config.add_simple_envs(self._supervisor_extra_env)
        supervisors_config.add_labels(self._supervisor_extra_labels)
        supervisors_config.create_namespaced(self._api_client, self._namespace)

    def _create_workers(self):
        workers_config = self._worker_config_cls(
            self._worker_num,
            image=self._image,
            cpu=self._worker_cpu,
            memory=self._worker_mem,
            memory_limit_ratio=self._worker_mem_limit_ratio,
            spill_volumes=self._worker_spill_paths,
            modules=self._worker_extra_modules,
            volumes=self._extra_volumes,
            worker_cache_mem=self._worker_cache_mem,
            min_cache_mem=self._worker_min_cache_men,
            service_name=self._service_name,
            service_port=self._worker_service_port,
            pre_stop_command=self._pre_stop_command,
            supervisor_web_port=self._web_port,
        )
        workers_config.add_simple_envs(self._worker_extra_env)
        workers_config.add_labels(self._worker_extra_labels)
        workers_config.create_namespaced(self._api_client, self._namespace)

    def _create_services(self):
        self._create_supervisors()
        self._create_workers()

    def _wait_services_ready(self):
        min_worker_num = int(self._min_worker_num or self._worker_num)
        limits = [self._supervisor_num, min_worker_num]
        selectors = [
            "mars/service-type=" + MarsSupervisorsConfig.rc_name,
            "mars/service-type=" + MarsWorkersConfig.rc_name,
        ]
        start_time = time.time()
        logger.debug("Start waiting pods to be ready")
        wait_services_ready(
            selectors,
            limits,
            lambda sel: self._get_ready_pod_count(sel),
            timeout=self._timeout,
        )
        logger.info("All service pods ready.")
        if self._timeout is not None:  # pragma: no branch
            self._timeout -= time.time() - start_time

    def _get_web_address(self):
        svc_data = self._core_api.read_namespaced_service(
            "marsservice", self._namespace
        ).to_dict()
        node_port = svc_data["spec"]["ports"][0]["node_port"]

        # docker desktop use a VM to hold docker processes, hence
        # we need to use API address instead
        desktop_nodes = self._core_api.list_node(
            field_selector="metadata.name=docker-desktop"
        ).to_dict()
        if desktop_nodes["items"]:  # pragma: no cover
            host_ip = urlparse(
                self._core_api.api_client.configuration.host
            ).netloc.split(":", 1)[0]
        else:
            web_pods = self._core_api.list_namespaced_pod(
                self._namespace,
                label_selector="mars/service-type=" + MarsSupervisorsConfig.rc_name,
            ).to_dict()
            host_ip = random.choice(web_pods["items"])["status"]["host_ip"]
        return f"http://{host_ip}:{node_port}"

    def _wait_web_ready(self):
        loop = new_isolation().loop

        async def get_supervisors():
            start_time = time.time()
            while True:
                try:
                    cluster_api = WebClusterAPI(self._external_web_endpoint)
                    supervisors = await cluster_api.get_supervisors()

                    if len(supervisors) == self._supervisor_num:
                        break
                except:  # noqa: E722  # nosec  # pylint: disable=bare-except  # pragma: no cover
                    if (
                        self._timeout is not None
                        and time.time() - start_time > self._timeout
                    ):
                        logger.exception("Error when fetching supervisors")
                        raise TimeoutError(
                            "Wait for kubernetes cluster timed out"
                        ) from None

        asyncio.run_coroutine_threadsafe(get_supervisors(), loop).result()

    def _load_cluster_logs(self):
        log_dict = dict()
        pod_items = self._core_api.list_namespaced_pod(self._namespace).to_dict()
        for item in pod_items["items"]:
            log_dict[item["metadata"]["name"]] = self._core_api.read_namespaced_pod_log(
                name=item["metadata"]["name"], namespace=self._namespace
            )
        return log_dict

    def start(self):
        try:
            self._create_namespace()
            self._create_roles_and_bindings()

            self._create_services()
            self._create_kube_service()

            self._wait_services_ready()

            self._external_web_endpoint = self._get_web_address()
            self._wait_web_ready()
            return self._external_web_endpoint
        except:  # noqa: E722
            if self._log_when_fail:  # pargma: no cover
                logger.error("Error when creating cluster")
                for name, log in self._load_cluster_logs().items():
                    logger.error("Error logs for %s:\n%s", name, log)
            self.stop()
            raise

    def stop(self, wait=False, timeout=0):
        # stop isolation
        stop_isolation()

        from kubernetes.client import CoreV1Api

        api = CoreV1Api(self._api_client)
        api.delete_namespace(self._namespace)
        if wait:
            start_time = time.time()
            while True:
                try:
                    api.read_namespace(self._namespace)
                except K8SApiException as ex:
                    if ex.status != 404:  # pragma: no cover
                        raise
                    break
                else:
                    time.sleep(1)
                    if (
                        timeout and time.time() - start_time > timeout
                    ):  # pragma: no cover
                        raise TimeoutError


def new_cluster(
    kube_api_client=None,
    image=None,
    supervisor_num=1,
    supervisor_cpu=None,
    supervisor_mem=None,
    worker_num=1,
    worker_cpu=None,
    worker_mem=None,
    worker_spill_paths=None,
    worker_cache_mem=None,
    min_worker_num=None,
    web_num=1,
    web_cpu=None,
    web_mem=None,
    service_type=None,
    timeout=None,
    **kwargs,
):
    """
    :param kube_api_client: Kubernetes API client, can be created with ``new_client_from_config``
    :param image: Docker image to use, ``marsproject/mars:<mars version>`` by default
    :param supervisor_num: Number of supervisors in the cluster, 1 by default
    :param supervisor_cpu: Number of CPUs for every supervisor
    :param supervisor_mem: Memory size for every supervisor
    :param worker_num: Number of workers in the cluster, 1 by default
    :param worker_cpu: Number of CPUs for every worker
    :param worker_mem: Memory size for every worker
    :param worker_spill_paths: Spill paths for worker pods on hosts
    :param worker_cache_mem: Size or ratio of cache memory for every worker
    :param min_worker_num: Minimal ready workers
    :param web_num: Number of web services in the cluster, 1 by default
    :param web_cpu: Number of CPUs for every web service
    :param web_mem: Memory size for every web service
    :param service_type: Type of Kubernetes Service, currently only ``NodePort`` supported
    :param timeout: Timeout when creating clusters
    """
    cluster_cls = kwargs.pop("cluster_cls", KubernetesCluster)
    cluster = cluster_cls(
        kube_api_client,
        image=image,
        supervisor_num=supervisor_num,
        supervisor_cpu=supervisor_cpu,
        supervisor_mem=supervisor_mem,
        worker_num=worker_num,
        worker_cpu=worker_cpu,
        worker_mem=worker_mem,
        worker_spill_paths=worker_spill_paths,
        worker_cache_mem=worker_cache_mem,
        min_worker_num=min_worker_num,
        web_num=web_num,
        web_cpu=web_cpu,
        web_mem=web_mem,
        service_type=service_type,
        timeout=timeout,
        **kwargs,
    )
    client = KubernetesClusterClient(cluster)
    client.start()
    return client



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ... import opcodes as OperandDef
from ...serialization.serializables import KeyField, ListField
from ..operands import TensorHasInput, TensorOperandMixin
from ..array_utils import get_array_module
from ..core import TensorOrder


class TensorSlice(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.SLICE

    _input = KeyField("input")
    _slices = ListField("slices")

    def __init__(self, slices=None, **kw):
        super().__init__(_slices=slices, **kw)

    @property
    def slices(self):
        return self._slices

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    def _get_order(self, kw, i):
        order = kw.pop("order", None)
        if order is None:
            inp = self.input
            if inp is None or inp.order == TensorOrder.C_ORDER:
                return TensorOrder.C_ORDER

            for shape, slc in zip(inp.shape, self._slices):
                if slc is None:
                    continue
                s = slc.indices(shape)
                if s[0] == 0 and s[1] == shape and s[2] == 1:
                    continue
                else:
                    return TensorOrder.C_ORDER

            return inp.order

        return order[i] if isinstance(order, (list, tuple)) else order

    @classmethod
    def execute(cls, ctx, op):
        inp = ctx[op.inputs[0].key]
        if op.input.ndim == 0 and not hasattr(inp, "shape"):
            # scalar, but organize it into an array
            inp = get_array_module(inp).array(inp)
        x = inp[tuple(op.slices)]
        out = op.outputs[0]
        ctx[out.key] = x.astype(x.dtype, order=out.order.value, copy=False)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from .manager import TaskConfigurationActor, TaskManagerActor


class TaskSupervisorService(AbstractService):
    """
    Task service on supervisor.

    Service Configuration
    ---------------------
    {
        "task": {
            "default_config": {
                "optimize_tileable_graph": True,
                "optimize_chunk_graph": True,
                "fuse_enabled": True
            },
            "execution_config": {
                "backend": "mars",
                "mars": {},
            }
        }
    }
    """

    async def start(self):
        task_config = self._config.get("task", dict())
        options = task_config.get("default_config", dict())
        execution_config = task_config.get("execution_config", dict())
        task_processor_cls = task_config.get("task_processor_cls")
        task_preprocessor_cls = task_config.get("task_preprocessor_cls")
        await mo.create_actor(
            TaskConfigurationActor,
            options,
            execution_config=execution_config,
            task_processor_cls=task_processor_cls,
            task_preprocessor_cls=task_preprocessor_cls,
            address=self._address,
            uid=TaskConfigurationActor.default_uid(),
        )

    async def stop(self):
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=TaskConfigurationActor.default_uid(), address=self._address
            )
        )

    async def create_session(self, session_id: str):
        await mo.create_actor(
            TaskManagerActor,
            session_id,
            address=self._address,
            uid=TaskManagerActor.gen_uid(session_id),
        )

    async def destroy_session(self, session_id: str):
        task_manager_ref = await mo.actor_ref(
            self._address, TaskManagerActor.gen_uid(session_id)
        )
        return await mo.destroy_actor(task_manager_ref)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from string import ascii_letters

from ...utils import tokenize
from ..operands import TensorFuse
from .. import arithmetic
from .core import TensorFuseChunkMixin


class TensorCpFuseChunk(TensorFuse, TensorFuseChunkMixin):
    # use for cupy-fused operand
    _op_type_ = None  # no opcode, cannot be serialized

    @classmethod
    def execute(cls, ctx, op):
        import cupy as cp

        chunk = op.outputs[0]
        func = cp.ElementwiseKernel(*_evaluate(chunk))
        ctx[chunk.key] = func(*[ctx[i.key] for i in op.inputs])


# execution part
CP_BINOP_TO_STRING = {
    arithmetic.TensorSubtract: "-",
    arithmetic.TensorMultiply: "*",
    arithmetic.TensorTrueDiv: "/",
}

CP_UNARYOP_TO_STRING = {
    arithmetic.TensorSqrt: "sqrt",
}


def _evaluate(chunk):
    letters = iter(letter for letter in ascii_letters if letter not in "ni")

    input_types = [i.dtype.name for i in chunk.op.inputs]
    input_names = {i: next(letters) for i in chunk.op.inputs}
    input_arguments = ", ".join(
        [f"{tp} {input_names[i]}" for i, tp in zip(chunk.op.inputs, input_types)]
    )
    output_type = chunk.op.dtype.name
    output_name = next(letters)
    output_argument = f"{output_type} {output_name}"
    body = dict(input_names)

    for node in chunk.composed:
        op_cls = type(node.op)
        if op_cls in CP_BINOP_TO_STRING:
            input_bodies = [body.get(i, repr(i)) for i in (node.op.lhs, node.op.rhs)]
            body[node] = f" {CP_BINOP_TO_STRING[op_cls]} ".join(input_bodies)
        elif op_cls in CP_UNARYOP_TO_STRING:
            input_data = body[node.op.inputs[0]]
            body[node] = f"{CP_UNARYOP_TO_STRING[op_cls]}({input_data})"
        else:
            raise NotImplementedError

    body = f"{output_name} = {body[chunk.composed[-1]]}"
    key = tokenize(input_arguments, output_argument, body)
    return (
        input_arguments,
        output_argument,
        body,
        f"{type(chunk.op).__name__.lower()}_{key}",
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import concurrent.futures

from .... import remote
from ....deploy.oscar.session import get_default_session, new_session

try:
    from joblib.parallel import (
        ParallelBackendBase,
        AutoBatchingMixin,
        register_parallel_backend,
    )
except ImportError:
    ParallelBackendBase = object
    AutoBatchingMixin = object
    register_parallel_backend = None


class MarsDistributedBackend(AutoBatchingMixin, ParallelBackendBase):
    MIN_IDEAL_BATCH_DURATION = 0.2
    MAX_IDEAL_BATCH_DURATION = 1.0
    supports_timeout = True

    def __init__(self, service=None, session=None, backend=None, n_parallel=None):
        super().__init__()

        if session is None:
            if service is not None:
                self.session = new_session(service, backend=backend, default=False)
            else:
                self.session = get_default_session()
        else:
            self.session = session

        self.n_parallel = n_parallel or 1
        self.executor = None

    def get_nested_backend(self):
        return MarsDistributedBackend(session=self.session), -1

    def configure(self, n_jobs=1, parallel=None, **backend_args):
        self.parallel = parallel
        n_parallel = self.effective_n_jobs(n_jobs)
        self.executor = concurrent.futures.ThreadPoolExecutor(n_parallel)
        return n_parallel

    def effective_n_jobs(self, n_jobs):
        eff_n_jobs = super(MarsDistributedBackend, self).effective_n_jobs(n_jobs)
        if n_jobs == -1:
            eff_n_jobs = self.n_parallel
        return eff_n_jobs

    def apply_async(self, func, callback=None):
        # todo allow execute f() in remote end to reduce data copy latency
        def f():
            spawned = []
            for func_obj, args, kwargs in func.items:
                spawned.append(remote.spawn(func_obj, args=args, kwargs=kwargs))

            ret = (
                remote.ExecutableTuple(spawned)
                .execute(session=self.session)
                .fetch(self.session)
            )
            callback(ret)
            return ret

        future = self.executor.submit(f)
        future.get = future.result
        return future


def register_mars_backend():
    register_parallel_backend("mars", MarsDistributedBackend)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio


class MutableTensorInfo:
    """
    Why `MutableTensorInfo`?

    We need a cluster to transfer meta information of mutable tensor, between
    server and client, as over the HTTP web session.

    Thus we design an internal-only `MutableTensorInfo` type as a container
    for those information.

    A `MutableTensor` can be initialized from

        - a info, which contains the metadata
        - a `mutable_api`, which will be used to request the backend API
        - a `loop`, which will be used to execute `__setitem__` (and `__getitem__`)
          synchronously to make the API more user-friendly.
    """

    def __init__(self, shape, dtype, name, default_value):
        self._shape = shape
        self._dtype = dtype
        self._name = name
        self._default_value = default_value

    @property
    def shape(self):
        return self._shape

    @property
    def dtype(self):
        return self._dtype

    @property
    def name(self):
        return self._name

    @property
    def default_value(self):
        return self._default_value


class MutableTensor:
    def __init__(self, info, mutable_api, loop):
        self._info = info
        self._mutable_api = mutable_api
        self._loop = loop

    @classmethod
    def create(
        cls,
        info: "MutableTensorInfo",
        mutable_api,  # no type signature, to avoid cycle imports
        loop: asyncio.AbstractEventLoop,
    ) -> "MutableTensor":
        return MutableTensor(info, mutable_api, loop)

    @property
    def shape(self):
        """
        Get the shape the mutable tensor.

        Returns
        -------
            Tuple
        """
        return self._info.shape

    @property
    def dtype(self):
        """
        Get the dtype the mutable tensor.

        Returns
        -------
            np.dtype or str
        """
        return self._info.dtype

    @property
    def name(self):
        """
        Get the dtype the mutable tensor.

        Returns
        -------
            str
        """
        return self._info.name

    @property
    def default_value(self):
        """
        Get the dtype the mutable tensor.

        Returns
        -------
            int or float
        """
        return self._info.default_value

    async def read(self, index, timestamp=None):
        """
        Read value from mutable tensor.

        Parameters
        ----------
        index:
            Index to read from the tensor.

        timestamp: optional
            Timestamp to read value that happened before then.
        """
        return await self._mutable_api.read(self.name, index, timestamp)

    async def write(self, index, value, timestamp=None):
        """
        Write value to mutable tensor.

        Parameters
        ----------
        index:
            Index to write to the tensor.

        value:
            The value that will be filled into the mutable tensor according to `index`.

        timestamp: optional
            Timestamp to associated with the newly touched value.
        """
        return await self._mutable_api.write(self.name, index, value, timestamp)

    def __getitem__(self, index):
        """
        Read value from mutable tensor with a synchronous API.

        Parameters
        ----------
        index:
            Index to read from the tensor.
        """
        coro = self.read(index)
        fut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return fut.result()

    def __setitem__(self, index, value):
        """
        Write value to mutable tensor with a synchronous API.

        Parameters
        ----------
        index:
            Index to write to the tensor.

        value:
            The value that will be filled into the mutable tensor according to `index`.
        """
        coro = self.write(index, value)
        fut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        return fut.result()

    async def seal(self, timestamp=None):
        """
        Seal the mutable tensor by name.

        Parameters
        ----------
        timestamp: optional
            Operations that happened before timestamp will be sealed, and later ones will be discard.

        Returns
        -------
            object
        """
        return await self._mutable_api.seal_mutable_tensor(self.name, timestamp)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorReduction, TensorReductionMixin


class TensorMin(TensorReduction, TensorReductionMixin):
    _op_type_ = OperandDef.MIN
    _func_name = "min"

    def __init__(self, axis=None, keepdims=None, combine_size=None, stage=None, **kw):
        stage = self._rewrite_stage(stage)
        super().__init__(
            _axis=axis,
            _keepdims=keepdims,
            _combine_size=combine_size,
            stage=stage,
            **kw
        )

    @classmethod
    def _is_sparse(cls, input_sparse, shape):
        if input_sparse and len(shape) > 0:
            return True
        return False


def min(a, axis=None, out=None, keepdims=None, combine_size=None):
    """
    Return the minimum of a tensor or minimum along an axis.

    Parameters
    ----------
    a : array_like
        Input data.
    axis : None or int or tuple of ints, optional
        Axis or axes along which to operate.  By default, flattened input is
        used.

        If this is a tuple of ints, the minimum is selected over multiple axes,
        instead of a single axis or all the axes as before.
    out : Tensor, optional
        Alternative output tensor in which to place the result.  Must
        be of the same shape and buffer length as the expected output.
        See `doc.ufuncs` (Section "Output arguments") for more details.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input tensor.

        If the default value is passed, then `keepdims` will not be
        passed through to the `amin` method of sub-classes of
        `Tensor`, however any non-default value will be.  If the
        sub-classes `sum` method does not implement `keepdims` any
        exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    amin : Tensor or scalar
        Minimum of `a`. If `axis` is None, the result is a scalar value.
        If `axis` is given, the result is an array of dimension
        ``a.ndim - 1``.

    See Also
    --------
    amax :
        The maximum value of a tensor along a given axis, propagating any NaNs.
    nanmin :
        The minimum value of a tensor along a given axis, ignoring any NaNs.
    minimum :
        Element-wise minimum of two tensors, propagating any NaNs.
    fmin :
        Element-wise minimum of two tensors, ignoring any NaNs.
    argmin :
        Return the indices of the minimum values.

    nanmax, maximum, fmax

    Notes
    -----
    NaN values are propagated, that is if at least one item is NaN, the
    corresponding min value will be NaN as well. To ignore NaN values
    (MATLAB behavior), please use nanmin.

    Don't use `amin` for element-wise comparison of 2 tensors; when
    ``a.shape[0]`` is 2, ``minimum(a[0], a[1])`` is faster than
    ``amin(a, axis=0)``.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> a = mt.arange(4).reshape((2,2))
    >>> a.execute()
    array([[0, 1],
           [2, 3]])
    >>> mt.amin(a).execute()           # Minimum of the flattened array
    0
    >>> mt.amin(a, axis=0).execute()   # Minima along the first axis
    array([0, 1])
    >>> mt.amin(a, axis=1).execute()   # Minima along the second axis
    array([0, 2])

    >>> b = mt.arange(5, dtype=float)
    >>> b[2] = mt.NaN
    >>> mt.amin(b).execute()
    nan
    >>> mt.nanmin(b).execute()
    0.0

    """
    a = astensor(a)
    op = TensorMin(
        axis=axis, dtype=a.dtype, keepdims=keepdims, combine_size=combine_size
    )
    return op(a, out=out)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes
from ... import tensor as mt
from ...core import OutputType, recursive_tile
from ...serialization.serializables import KeyField
from ...tensor.array_utils import as_same_device, device, sparse
from ...tensor.core import TensorOrder
from ...utils import has_unknown_shape
from ..operands import LearnOperand, LearnOperandMixin
from ._k_means_fast import _inertia_dense, _inertia_sparse, merge_update_chunks


class KMeansInertia(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_INERTIA

    _x = KeyField("x")
    _sample_weight = KeyField("sample_weight")
    _centers = KeyField("centers")
    _labels = KeyField("labels")

    def __init__(
        self,
        x=None,
        sample_weight=None,
        centers=None,
        labels=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _x=x,
            _sample_weight=sample_weight,
            _centers=centers,
            _labels=labels,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor]

    @property
    def x(self):
        return self._x

    @property
    def sample_weight(self):
        return self._sample_weight

    @property
    def centers(self):
        return self._centers

    @property
    def labels(self):
        return self._labels

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in ("_x", "_sample_weight", "_centers", "_labels"):
            if getattr(self, field, None) is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        params = {"shape": (), "dtype": np.dtype(float), "order": TensorOrder.C_ORDER}
        return self.new_tileable(
            [self._x, self._sample_weight, self._centers, self._labels], kws=[params]
        )

    @classmethod
    def tile(cls, op: "KMeansInertia"):
        if has_unknown_shape(*op.inputs):
            yield
        x = op.x
        x = yield from recursive_tile(x.rechunk({1: x.shape[1]}))
        sample_weight = yield from recursive_tile(
            op.sample_weight.rechunk({0: x.nsplits[0]})
        )
        labels = yield from recursive_tile(op.labels.rechunk({0: x.nsplits[0]}))
        centers = op.centers
        centers = yield from recursive_tile(centers.rechunk(centers.shape))

        out_chunks = []
        for x_chunk, sample_weight_chunk, labels_chunk in zip(
            x.chunks, sample_weight.chunks, labels.chunks
        ):
            chunk_op = op.copy().reset_key()
            chunk_params = {
                "shape": (1,),
                "dtype": np.dtype(float),
                "order": TensorOrder.C_ORDER,
                "index": x_chunk.index,
            }
            out_chunk = chunk_op.new_chunk(
                [x_chunk, sample_weight_chunk, centers.chunks[0], labels_chunk],
                kws=[chunk_params],
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = op.outputs[0].params
        params["shape"] = (x.chunk_shape[0],)
        params["chunks"] = out_chunks
        params["nsplits"] = ((1,) * x.chunk_shape[0],)
        out = new_op.new_tileable(op.inputs, kws=[params]).sum()
        out = yield from recursive_tile(out)
        return [out]

    @classmethod
    def execute(cls, ctx, op):
        (x, sample_weight, centers, labels), device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs],
            device=op.device,
            ret_extra=True,
            copy_if_not_writeable=True,
        )

        with device(device_id):
            if xp is np:
                method = _inertia_dense
            elif xp is sparse:
                method = _inertia_sparse
            else:  # pragma: no cover
                raise NotImplementedError("Cannot run inertial on GPU")

            result = method(x, sample_weight, centers, labels)
            ctx[op.outputs[0].key] = np.array([result])


def _inertia(X, sample_weight, centers, labels):
    op = KMeansInertia(x=X, sample_weight=sample_weight, centers=centers, labels=labels)
    return op()


def _execute_merge_update(ctx, op):
    inputs, device_id, xp = as_same_device(
        [ctx[inp.key] for inp in op.inputs],
        op.device,
        ret_extra=True,
        copy_if_not_writeable=True,
    )
    length = len(inputs) // 2
    assert len(inputs) % 2 == 0
    centers_new_chunks = inputs[:length]
    weight_in_cluster_chunks = inputs[length:]

    with device(device_id):
        weight_in_clusters = np.zeros(
            op.n_clusters, dtype=weight_in_cluster_chunks[0].dtype
        )
        centers_new = np.zeros_like(centers_new_chunks[0])
        n_clusters = op.n_clusters
        n_features = centers_new_chunks[0].shape[1]

        for weight_in_clusters_chunk, centers_new_chunk in zip(
            weight_in_cluster_chunks, centers_new_chunks
        ):
            merge_update_chunks(
                n_clusters,
                n_features,
                weight_in_clusters,
                weight_in_clusters_chunk,
                centers_new,
                centers_new_chunk,
            )

        # centers new
        ctx[op.outputs[0].key] = centers_new
        # weight_in_clusters
        ctx[op.outputs[1].key] = weight_in_clusters


class KMeansRelocateEmptyClusters(LearnOperand, LearnOperandMixin):
    _op_type_ = opcodes.KMEANS_RELOCASTE_EMPTY_CLUSTERS

    _empty_clusters = KeyField("empty_clusters")
    _far_x = KeyField("far_x")
    _far_labels = KeyField("far_labels")
    _far_sample_weights = KeyField("far_sample_weight")
    _centers_new = KeyField("centers_new")
    _weight_in_clusters = KeyField("weight_in_clusters")

    def __init__(
        self,
        empty_clusters=None,
        far_x=None,
        far_labels=None,
        far_sample_weights=None,
        centers_new=None,
        weight_in_clusters=None,
        output_types=None,
        **kw
    ):
        super().__init__(
            _empty_clusters=empty_clusters,
            _far_x=far_x,
            _far_labels=far_labels,
            _far_sample_weights=far_sample_weights,
            _centers_new=centers_new,
            _weight_in_clusters=weight_in_clusters,
            _output_types=output_types,
            **kw
        )
        if self._output_types is None:
            self._output_types = [OutputType.tensor] * self.output_limit

    @property
    def empty_clusters(self):
        return self._empty_clusters

    @property
    def far_x(self):
        return self._far_x

    @property
    def far_labels(self):
        return self._far_labels

    @property
    def far_sample_weights(self):
        return self._far_sample_weights

    @property
    def centers_new(self):
        return self._centers_new

    @property
    def weight_in_clusters(self):
        return self._weight_in_clusters

    @property
    def output_limit(self):
        return 2

    @property
    def _input_fields(self):
        return (
            "_empty_clusters",
            "_far_x",
            "_far_labels",
            "_far_sample_weights",
            "_centers_new",
            "_weight_in_clusters",
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        for field in self._input_fields:
            ob = getattr(self, field)
            if ob is not None:
                setattr(self, field, next(inputs_iter))

    def __call__(self):
        kws = [
            # centers_new
            self._centers_new.params,
            # weight_in_clusters
            self._weight_in_clusters.params,
        ]
        return self.new_tileables(
            [getattr(self, field) for field in self._input_fields], kws=kws
        )

    @classmethod
    def tile(cls, op: "KMeansRelocateEmptyClusters"):
        empty_clusters = yield from recursive_tile(
            op.empty_clusters.rechunk(op.empty_clusters.shape)
        )
        far_x = yield from recursive_tile(op.far_x.rechunk(op.far_x.shape))
        far_labels = yield from recursive_tile(
            op.far_labels.rechunk(op.far_labels.shape)
        )
        far_sample_weight = yield from recursive_tile(
            op.far_sample_weights.rechunk(op.far_sample_weights.shape)
        )
        centers_new = yield from recursive_tile(
            op.centers_new.rechunk(op.centers_new.shape)
        )
        weight_in_clusters = yield from recursive_tile(
            op.weight_in_clusters.rechunk(op.weight_in_clusters.shape)
        )

        chunk_op = op.copy().reset_key()
        out_centers_new_chunk, out_weight_in_clusters_chunk = chunk_op.new_chunks(
            [
                empty_clusters.chunks[0],
                far_x.chunks[0],
                far_labels.chunks[0],
                far_sample_weight.chunks[0],
                centers_new.chunks[0],
                weight_in_clusters.chunks[0],
            ],
            kws=[centers_new.chunks[0].params, weight_in_clusters.chunks[0].params],
        )

        out_centers_new_params = centers_new.params
        out_centers_new_params["nsplits"] = centers_new.nsplits
        out_centers_new_params["chunks"] = [out_centers_new_chunk]
        out_weight_in_clusters_params = weight_in_clusters.params
        out_weight_in_clusters_params["nsplits"] = weight_in_clusters.nsplits
        out_weight_in_clusters_params["chunks"] = [out_weight_in_clusters_chunk]
        new_op = op.copy()
        return new_op.new_tileables(
            op.inputs, kws=[out_centers_new_params, out_weight_in_clusters_params]
        )

    @classmethod
    def execute(cls, ctx, op):
        (
            (
                empty_clusters,
                far_x,
                far_labels,
                far_sample_weight,
                center_new,
                weight_in_clusters,
            ),
            device_id,
            xp,
        ) = as_same_device(
            [ctx[inp.key] for inp in op.inputs], op.device, ret_extra=True
        )

        out_centers_new = center_new.copy()
        out_weight_in_clusters = weight_in_clusters.copy()
        del center_new, weight_in_clusters

        n_empty = empty_clusters.shape[0]
        n_features = far_x.shape[1]

        for idx in range(n_empty):
            new_cluster_id = empty_clusters[idx]
            weight = far_sample_weight[idx]
            old_cluster_id = far_labels[idx]

            for k in range(n_features):
                out_centers_new[old_cluster_id, k] -= far_x[idx, k] * weight
                out_centers_new[new_cluster_id, k] = far_x[idx, k] * weight

            out_weight_in_clusters[new_cluster_id] = weight
            out_weight_in_clusters[old_cluster_id] -= weight

        ctx[op.outputs[0].key] = out_centers_new
        ctx[op.outputs[1].key] = out_weight_in_clusters


def _relocate_empty_clusters(
    X,
    sample_weight,
    centers_old,
    centers_new,
    weight_in_clusters,
    labels,
    to_run=None,
    session=None,
    run_kwargs=None,
):
    to_run = to_run or list()
    empty_clusters = mt.where(mt.equal(weight_in_clusters, 0))[0].astype(mt.int32)
    to_run.append(empty_clusters)

    mt.ExecutableTuple(to_run).execute(session=session, **(run_kwargs or dict()))

    n_empty = empty_clusters.shape[0]

    if n_empty == 0:
        return centers_new, weight_in_clusters

    distances = ((mt.asarray(X) - mt.asarray(centers_old)[labels]) ** 2).sum(axis=1)
    far_from_centers = mt.argpartition(distances, -n_empty)[: -n_empty - 1 : -1].astype(
        np.int32
    )

    far_x = X[far_from_centers]
    far_labels = labels[far_from_centers]
    far_sample_weight = sample_weight[far_from_centers]

    op = KMeansRelocateEmptyClusters(
        empty_clusters=empty_clusters,
        far_x=far_x,
        far_labels=far_labels,
        far_sample_weights=far_sample_weight,
        centers_new=centers_new,
        weight_in_clusters=weight_in_clusters,
    )
    return op()



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, Type, Any

from .....resource import Resource


_name_to_resource_evaluator: Dict[str, Type["ResourceEvaluator"]] = {}


def register_resource_evaluator(evaluator_cls: Type["ResourceEvaluator"]):
    _name_to_resource_evaluator[evaluator_cls.name] = evaluator_cls
    return evaluator_cls


def init_default_resource_for_subtask(subtask_graph: "SubtaskGraph"):  # noqa: F821
    for subtask in subtask_graph.iter_nodes():
        is_gpu = any(c.op.gpu for c in subtask.chunk_graph)
        subtask.required_resource = (
            Resource(num_gpus=1) if is_gpu else Resource(num_cpus=1)
        )


class ResourceEvaluator(ABC):
    """
    Resource evaluator is used to estimate and set resources required by
    subtasks. It can be an internal service or an external service. If it
    is an internal service, we can set default of adjustable resources for
    subtasks. If it is an external service, we should report the running
    result of the task to the external service, so that it can accurately
    predict the required resources of subtasks based on the historical
    running information, we call it HBO.

    Best practice
    ----------
    You can follow the steps below to add a new resource evaluator:
        * Inherit `ResourceEvaluator` and implement `create`, `evaluate`
          and `report` methods. The `create` method is to create a new
          resource evaluator instance. The `evaluate` method is to estimate
          and set required resources for the subtasks of a task stage. And
          this method must be implemented. The `report` method is to report
          the running information and result of the task. And this method
          does not have to be implemented.

        * Add default configs of the new evaluator needed in `base_config.xml`
          or its descendant files.

        * Set the `resource_evaluator` to choose a resource evaluator in
          `base_config.xml` when running a mars job.
    """

    name = None

    @classmethod
    @abstractmethod
    async def create(cls, config: Dict[str, Any], **kwargs) -> "ResourceEvaluator":
        name = config.get("resource_evaluator", "default")
        evaluator_config = config.get(name, {})
        evaluator_cls = _name_to_resource_evaluator[name]
        return await evaluator_cls.create(evaluator_config, **kwargs)

    @abstractmethod
    async def evaluate(self, stage_processor: "TaskStageProcessor"):  # noqa: F821
        """Called before executing a task stage."""

    @abstractmethod
    async def report(self):
        """Called after executing a task."""


@register_resource_evaluator
class DefaultEvaluator(ResourceEvaluator):
    name = "default"

    @classmethod
    async def create(cls, config, **kwargs) -> "ResourceEvaluator":
        return cls()

    async def evaluate(self, stage_processor: "TaskStageProcessor"):  # noqa: F821
        init_default_resource_for_subtask(stage_processor.subtask_graph)

    async def report(self):
        pass



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This utility module dumps code of remote traceback and loads them
into local linecache. This enables displaying codes of remote
tracebacks correctly.
"""

import linecache
import os
import types
from collections import defaultdict


def dump_traceback_code(tb: types.TracebackType, number_of_lines_of_context: int = 5):
    """
    Dump codes before and after lines of tracebacks.

    Parameters
    ----------
    tb: types.TracebackType
        Traceback object
    number_of_lines_of_context: int
        Total number of lines around the code
    Returns
    -------
    result: dict
        Dumped code lines of traceback
    """
    results = defaultdict(lambda: dict(fragments=[]))

    while tb:
        file_name = tb.tb_frame.f_code.co_filename
        if linecache.getline(file_name, tb.tb_lineno):  # pragma: no branch
            code_lines = linecache.cache[file_name][2]
            left_range = max(tb.tb_lineno - number_of_lines_of_context // 2 - 1, 0)
            right_range = min(left_range + number_of_lines_of_context, len(code_lines))

            cache_data = linecache.cache[file_name]
            fragment = cache_data[2][left_range:right_range]
            results[file_name]["fragments"].append(
                dict(left=left_range, right=right_range, code=fragment)
            )
            results[file_name].update(
                dict(size=cache_data[0], lines=len(cache_data[2]))
            )
        tb = tb.tb_next
    return dict(results)


def load_traceback_code(code_frags: dict, cache: dict = None):
    """
    Load dumped codes for remote tracebacks.

    Parameters
    ----------
    code_frags: dict
        Dumped codes for remote traceback.
    cache: dict
        Target for codes to be dumped, for test purpose only.
        Production code should keep this field as None.
    """
    if cache is not None:
        real_cache = False
    else:
        real_cache = True
        cache = linecache.cache

    for file_name, profile in code_frags.items():
        if real_cache and os.path.exists(file_name):
            # skip rewriting caches of existing files
            continue

        if file_name not in cache:
            # keep field 1 (mtime) as None to ensure lazy cache
            cache[file_name] = (
                profile["size"],
                None,
                [""] * profile["lines"],
                file_name,
            )
        for fragment in profile["fragments"]:
            left_range, right_range = fragment["left"], fragment["right"]
            cache[file_name][2][left_range:right_range] = fragment["code"]



# Copyright 1999-2020 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType
from ...serialization.serializables import FieldTypes, StringField, TupleField
from ...tensor.datastore.to_vineyard import resolve_vineyard_socket
from ...utils import lazy_import
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index

vineyard = lazy_import("vineyard")
vy_data_df = lazy_import("vineyard.data.dataframe", rename="vy_data_df")
vy_data_utils = lazy_import("vineyard.data.utils", rename="vy_data_utils")


class DataFrameToVineyardChunk(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_CHUNK

    # vineyard ipc socket
    vineyard_socket = StringField("vineyard_socket")

    # a dummy attr to make sure ops have different keys
    operator_index = TupleField("operator_index", FieldTypes.int32)

    def __init__(self, vineyard_socket=None, dtypes=None, **kw):
        super().__init__(
            vineyard_socket=vineyard_socket,
            _dtypes=dtypes,
            _output_types=[OutputType.dataframe],
            **kw
        )

    def __call__(self, df):
        return self.new_dataframe(
            [df],
            shape=(0, 0),
            dtypes=df.dtypes,
            index_value=df.index_value,
            columns_value=df.columns_value,
        )

    @classmethod
    def _process_out_chunks(cls, op, out_chunks):
        dtypes = pd.Series([np.dtype("O")], index=pd.Index([0]))
        merge_op = DataFrameToVinyardStoreMeta(
            vineyard_socket=op.vineyard_socket,
            chunk_shape=op.inputs[0].chunk_shape,
            shape=(1, 1),
            dtypes=dtypes,
        )
        return merge_op.new_chunks(
            out_chunks, shape=(1, 1), dtypes=dtypes, index=(0, 0)
        )

    @classmethod
    def tile(cls, op):
        out_chunks = []
        dtypes = pd.Series([np.dtype("O")], index=pd.Index([0]))
        for idx, chunk in enumerate(op.inputs[0].chunks):
            chunk_op = op.copy().reset_key()
            chunk_op.operator_index = chunk.index
            out_chunk = chunk_op.new_chunk(
                [chunk],
                shape=(1, 1),
                dtypes=dtypes,
                index_value=chunk.index_value,
                columns_value=chunk.columns_value,
                index=(idx, 0),
            )
            out_chunks.append(out_chunk)
        out_chunks = cls._process_out_chunks(op, out_chunks)

        in_df = op.inputs[0]
        new_op = op.copy().reset_key()
        return new_op.new_dataframes(
            op.inputs,
            shape=(len(out_chunks), 1),
            dtypes=dtypes,
            index_value=in_df.index_value,
            columns_value=in_df.columns_value,
            chunks=out_chunks,
            nsplits=((np.prod(op.inputs[0].chunk_shape),),),
        )

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError("vineyard is not available")

        socket, needs_put = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        # some op might be fused and executed twice on different workers
        if not needs_put:
            # might be fused
            try:  # pragma: no cover
                meta = ctx.get_chunks_meta([op.inputs[0].key])[0]
                df_id = vineyard.ObjectID(meta["object_ref"])
                if not client.exists(df_id):
                    needs_put = True
            except KeyError:
                needs_put = True
        if needs_put:
            df_id = client.put(
                ctx[op.inputs[0].key], partition_index=op.inputs[0].index
            )
        else:  # pragma: no cover
            meta = client.get_meta(df_id)
            new_meta = vineyard.ObjectMeta()
            for k, v in meta.items():
                if k not in ["id", "signature", "instance_id"]:
                    if isinstance(v, vineyard.ObjectMeta):
                        new_meta.add_member(k, v)
                    else:
                        new_meta[k] = v
            new_meta["partition_index_"] = vy_data_utils.to_json(op.inputs[0].index)
            df_id = client.create_metadata(new_meta).id

        client.persist(df_id)
        ctx[op.outputs[0].key] = pd.DataFrame({0: [df_id]})


class DataFrameToVinyardStoreMeta(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_STORE_VINEYARD_META

    # vineyard ipc socket
    vineyard_socket = StringField("vineyard_socket")

    def __init__(self, vineyard_socket=None, dtypes=None, **kw):
        super().__init__(
            vineyard_socket=vineyard_socket,
            dtypes=dtypes,
            _output_types=[OutputType.dataframe],
            **kw
        )

    @classmethod
    def tile(cls, op):
        dtypes = pd.Series([np.dtype("O")], index=pd.Index([0]))
        chunk_op = op.copy().reset_key()
        out_chunk = chunk_op.new_chunk(
            op.inputs[0].chunks,
            shape=(1, 1),
            dtypes=dtypes,
            index_value=parse_index(pd.Index([-1])),
            columns_value=parse_index(pd.Index([0])),
            index=(0, 0),
        )
        new_op = op.copy().reset_key()
        return new_op.new_dataframes(
            op.inputs,
            shape=(1, 1),
            dtypes=dtypes,
            index_value=parse_index(pd.Index([0])),
            columns_value=parse_index(pd.Index([0])),
            chunks=[out_chunk],
            nsplits=((1,), (1,)),
        )

    @classmethod
    def execute(cls, ctx, op):
        if vineyard is None:
            raise RuntimeError("vineyard is not available")

        socket, _ = resolve_vineyard_socket(ctx, op)
        client = vineyard.connect(socket)

        # # store the result object id to execution context
        chunks = [ctx[chunk.key][0][0] for chunk in op.inputs]
        ctx[op.outputs[0].key] = pd.DataFrame(
            {0: [vy_data_df.make_global_dataframe(client, chunks).id]}
        )


def to_vineyard(df, vineyard_socket=None):
    op = DataFrameToVineyardChunk(vineyard_socket=vineyard_socket)
    return op(df)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from ....serialization.serializables import (
    BoolField,
    Int32Field,
    Int64Field,
    StringField,
)
from ....utils import pd_release_version
from ...utils import validate_axis
from ..core import Window

_window_has_method = pd_release_version >= (1, 3, 0)


class Expanding(Window):
    _min_periods = Int64Field("min_periods")
    _axis = Int32Field("axis")
    _center = BoolField("center")
    _method = StringField("method")

    def __init__(self, min_periods=None, axis=None, center=None, method=None, **kw):
        super().__init__(
            _min_periods=min_periods, _axis=axis, _center=center, _method=method, **kw
        )

    @property
    def min_periods(self):
        return self._min_periods

    @property
    def axis(self):
        return self._axis

    @property
    def center(self):
        return self._center

    @property
    def method(self):
        return self._method or "single"

    def __call__(self, df):
        return df.expanding(**self.params)

    @property
    def params(self):
        p = OrderedDict()

        if not _window_has_method:  # pragma: no cover
            args = ["min_periods", "center", "axis"]
        else:
            args = ["min_periods", "center", "axis", "method"]

        for k in args:
            p[k] = getattr(self, k)
        return p

    def aggregate(self, func, **kwargs):
        from .aggregation import DataFrameExpandingAgg

        count_always_valid = kwargs.pop("_count_always_valid", False)

        op = DataFrameExpandingAgg(
            func=func, count_always_valid=count_always_valid, **self.params
        )
        return op(self)

    agg = aggregate

    def sum(self):
        return self.aggregate("sum")

    def count(self):
        return self.aggregate("count")

    def min(self):
        return self.aggregate("min")

    def max(self):
        return self.aggregate("max")

    def mean(self):
        return self.aggregate("mean")

    def var(self):
        return self.aggregate("var")

    def std(self):
        return self.aggregate("std")


def expanding(obj, min_periods=1, center=False, axis=0):
    """
    Provide expanding transformations.

    Parameters
    ----------
    min_periods : int, default 1
    Minimum number of observations in window required to have a value
    (otherwise result is NA).
    center : bool, default False
    Set the labels at the center of the window.
    axis : int or str, default 0

    Returns
    -------
    a Window sub-classed for the particular operation

    See Also
    --------
    rolling : Provides rolling window calculations.
    ewm : Provides exponential weighted functions.

    Notes
    -----
    By default, the result is set to the right edge of the window. This can be
    changed to the center of the window by setting ``center=True``.

    Examples
    --------
    >>> import numpy as np
    >>> import mars.dataframe as md
    >>> df = md.DataFrame({'B': [0, 1, 2, np.nan, 4]})
    >>> df.execute()
         B
    0  0.0
    1  1.0
    2  2.0
    3  NaN
    4  4.0
    >>> df.expanding(2).sum().execute()
         B
    0  NaN
    1  1.0
    2  3.0
    3  3.0
    4  7.0
    """
    axis = validate_axis(axis, obj)

    if center:
        raise NotImplementedError("center == True is not supported")
    if axis == 1:
        raise NotImplementedError("axis other than 0 is not supported")

    return Expanding(input=obj, min_periods=min_periods, center=center, axis=axis)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import List, Union

from ....typing import TileableType
from ...mode import enter_mode
from ..entity import TileableGraph, ChunkGraph
from .tileable import TileableGraphBuilder
from .chunk import ChunkGraphBuilder


@enter_mode(kernel=True)
def build_graph(
    tileables: List[TileableType],
    tile: bool = False,
    fuse_enabled: bool = True,
    **chunk_graph_build_kwargs
) -> Union[TileableGraph, ChunkGraph]:
    tileables = list(itertools.chain(*(tileable.op.outputs for tileable in tileables)))
    tileable_graph = TileableGraph(tileables)
    tileable_graph_builder = TileableGraphBuilder(tileable_graph)
    tileable_graph = next(tileable_graph_builder.build())
    if not tile:
        return tileable_graph
    chunk_graph_builder = ChunkGraphBuilder(
        tileable_graph, fuse_enabled=fuse_enabled, **chunk_graph_build_kwargs
    )
    return next(chunk_graph_builder.build())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral

import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE
from ...core.operand import OperandStage
from ...serialization.serializables import (
    FieldTypes,
    KeyField,
    ListField,
    TupleField,
    Int32Field,
)
from ..core import TENSOR_TYPE, TensorOrder
from ..utils import split_indexes_into_chunks, calc_pos, filter_inputs
from ..operands import TensorHasInput, TensorOperandMixin, TensorMapReduceOperand
from ..array_utils import get_array_module
from .core import process_index, calc_shape
from .index_lib import TensorIndexesHandler

FANCY_INDEX_TYPES = TENSOR_TYPE + (np.ndarray,)


class TensorIndex(TensorHasInput, TensorOperandMixin):
    _op_type_ = OperandDef.INDEX

    _input = KeyField("input")
    _indexes = ListField("indexes")

    def __init__(self, indexes=None, **kw):
        super().__init__(_indexes=indexes, **kw)

    @property
    def indexes(self):
        return self._indexes

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs[1:])
        new_indexes = [
            next(inputs_iter) if isinstance(index, ENTITY_TYPE) else index
            for index in self._indexes
        ]
        self._indexes = new_indexes

    def on_output_modify(self, new_output):
        from .setitem import TensorIndexSetValue

        if self.create_view:
            a = self.input
            op = TensorIndexSetValue(
                dtype=a.dtype,
                sparse=a.issparse(),
                indexes=tuple(self._indexes),
                value=new_output,
            )
            return op(a, self._indexes, new_output)

    def on_input_modify(self, new_input):
        if self.create_view:
            new_op = self.copy().reset_key()
            new_inputs = [new_input] + self.inputs[1:]
            return new_op.new_tensor(new_inputs, shape=self.outputs[0].shape)

    def __call__(self, a, index, shape, order):
        self._indexes = list(index)
        return self.new_tensor(filter_inputs([a] + list(index)), shape, order=order)

    @classmethod
    def _tile_one_chunk(cls, op: "TensorIndex"):
        inp = op.inputs[0]
        out = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_params = out.params.copy()
        chunk_params["shape"] = shape = tuple(calc_shape(inp.shape, op.indexes))
        chunk_params["index"] = (0,) * out.ndim
        chunk = chunk_op.new_chunk(
            [inp.chunks[0] for inp in op.inputs], kws=[chunk_params]
        )
        params = out.params.copy()
        params["chunks"] = [chunk]
        params["nsplits"] = tuple((s,) for s in shape)
        return op.copy().new_tensors(op.inputs, kws=[params])

    @classmethod
    def tile(cls, op):
        if all(len(inp.chunks) == 1 for inp in op.inputs):
            return cls._tile_one_chunk(op)

        handler = TensorIndexesHandler()
        return [(yield from handler.handle(op))]

    @classmethod
    def execute(cls, ctx, op):
        indexes = tuple(
            ctx[index.key] if hasattr(index, "key") else index for index in op.indexes
        )
        input_ = ctx[op.inputs[0].key]
        xp = get_array_module(input_)
        ret = xp.asarray(input_)[indexes]
        if hasattr(ret, "astype"):
            ret = ret.astype(ret.dtype, order=op.outputs[0].order.value, copy=False)
        ctx[op.outputs[0].key] = ret

    @classmethod
    def estimate_size(cls, ctx, op):
        chunk = op.outputs[0]
        shape = chunk.shape

        if any(np.isnan(s) for s in shape):
            return super().estimate_size(ctx, op)

        new_indexes = [index for index in op._indexes if index is not None]
        new_shape = []
        first_fancy_index = False
        for index in new_indexes:
            if isinstance(index, ENTITY_TYPE):
                if index.dtype != np.bool_:
                    if not first_fancy_index:
                        first_fancy_index = True
                    else:
                        continue
                new_shape.append(ctx[index.key][0] // index.dtype.itemsize)

        rough_shape = []
        idx = 0
        for s in shape:
            if np.isnan(s):
                rough_shape.append(new_shape[idx])
                idx += 1
            else:
                rough_shape.append(s)
        result = int(np.prod(rough_shape) * chunk.dtype.itemsize)
        ctx[chunk.key] = (result, result)


class FancyIndexingDistribute(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.FANCY_INDEX_DISTRIBUTE

    _input = KeyField("input")
    _dest_nsplits = TupleField("dest_nsplits", FieldTypes.tuple(FieldTypes.uint64))
    _axes = TupleField("axes", FieldTypes.int32)

    def __init__(self, dest_nsplits=None, axes=None, **kw):
        super().__init__(_dest_nsplits=dest_nsplits, _axes=axes, **kw)

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]

    @property
    def output_limit(self):
        if self.stage == OperandStage.map:
            return 1
        # return fancy indexes on each axis as well as original position
        return len(self._axes) + 1

    @property
    def dest_nsplits(self):
        return self._dest_nsplits

    @property
    def axes(self):
        return self._axes

    @classmethod
    def _execute_map(cls, ctx, op):
        nsplits = op.dest_nsplits
        axes = op.axes
        fancy_index_nsplits = [nsplits[ax] for ax in axes]
        indexes = ctx[op.inputs[0].key]
        flatten_indexes = indexes.reshape(indexes.shape[0], -1)
        idx_to_fancy_indexes, idx_to_poses = split_indexes_into_chunks(
            fancy_index_nsplits, flatten_indexes, False
        )
        for idx in idx_to_fancy_indexes:
            ctx[op.outputs[0].key, idx] = (idx_to_fancy_indexes[idx], idx_to_poses[idx])

    @classmethod
    def _execute_reduce(cls, ctx, op: "FancyIndexingDistribute"):
        fancy_indexes = []
        poses = []
        xp = None
        for fancy_index, pos in op.iter_mapper_data(ctx):
            if xp is None:
                xp = get_array_module(fancy_index)
            if fancy_index.size == 0:
                fancy_index = fancy_index.reshape(len(op.axes), 0)
            fancy_indexes.append(fancy_index)
            poses.append(pos)

        fancy_index = np.hstack(fancy_indexes)
        pos = np.hstack(poses)

        assert len(op.outputs) - 1 == len(fancy_index)
        for out_chunk, axis_fancy_index in zip(op.outputs[:-1], fancy_index):
            ctx[out_chunk.key] = axis_fancy_index
        ctx[op.outputs[-1].key] = np.asarray([len(p) for p in poses]), pos

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)

    @classmethod
    def estimate_size(cls, ctx, op):
        if op.stage == OperandStage.map:
            fancy_index_size = len(op.axes)
            inp_size = ctx[op.inputs[0].key][0]
            factor = (
                1 / float(fancy_index_size) + fancy_index_size
            )  # 1/#fancy_index is the poses
            ctx[op.outputs[0].key] = (inp_size * factor,) * 2
        else:
            sum_size = 0
            for shuffle_input in op.inputs[0].inputs or ():
                sum_size += ctx[shuffle_input.key]
            for out_chunk in op.outputs:
                ctx[out_chunk.key] = sum_size, sum_size


class FancyIndexingConcat(TensorMapReduceOperand, TensorOperandMixin):
    _op_type_ = OperandDef.FANCY_INDEX_CONCAT

    _fancy_index_axis = Int32Field("fancy_index_axis")
    _fancy_index_shape = TupleField("fancy_index_shape", FieldTypes.int64)

    def __init__(self, fancy_index_axis=None, fancy_index_shape=None, **kw):
        super().__init__(
            _fancy_index_axis=fancy_index_axis,
            _fancy_index_shape=fancy_index_shape,
            **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def fancy_index_axis(self):
        return self._fancy_index_axis

    @property
    def fancy_index_shape(self):
        return self._fancy_index_shape

    @classmethod
    def _execute_map(cls, ctx, op):
        indexed_array = ctx[op.inputs[0].key]
        sizes, pos = ctx[op.inputs[1].key]
        acc_sizes = np.cumsum(sizes)
        fancy_index_axis = op.fancy_index_axis

        for i in range(len(sizes)):
            start = 0 if i == 0 else acc_sizes[i - 1]
            end = acc_sizes[i]
            select = (slice(None),) * fancy_index_axis + (slice(start, end),)
            ctx[op.outputs[0].key, (i,)] = (indexed_array[select], pos[start:end])

    @classmethod
    def _execute_reduce(cls, ctx, op: "FancyIndexingConcat"):
        fancy_index_axis = op.fancy_index_axis
        fancy_index_shape = op.fancy_index_shape

        indexed_arrays = []
        poses = []
        for index_array, pos in op.iter_mapper_data(ctx):
            indexed_arrays.append(index_array)
            poses.append(pos)

        concat_array = get_array_module(indexed_arrays[0]).concatenate(
            indexed_arrays, axis=fancy_index_axis
        )
        concat_pos = get_array_module(poses[0]).hstack(poses)
        select_pos = calc_pos(
            fancy_index_shape, concat_pos, xp=get_array_module(poses[0])
        )
        select = (slice(None),) * fancy_index_axis + (select_pos,)
        ctx[op.outputs[0].key] = concat_array[select]

    @classmethod
    def execute(cls, ctx, op):
        if op.stage == OperandStage.map:
            cls._execute_map(ctx, op)
        else:
            cls._execute_reduce(ctx, op)

    @classmethod
    def estimate_size(cls, ctx, op):
        if op.stage == OperandStage.map:
            input_size = ctx[op.inputs[0].key][0]
            pos_size = ctx[op.inputs[0].key][0]
            ctx[op.outputs[0].key] = input_size + pos_size, input_size + pos_size * 2
        else:
            chunk = op.outputs[0]
            input_sizes = [ctx[c.key][0] for c in op.inputs[0].inputs or ()]
            ctx[chunk.key] = chunk.nbytes, chunk.nbytes + sum(input_sizes)


def _is_bool_index(index_obj):
    return isinstance(index_obj, TENSOR_TYPE) and index_obj.dtype == np.bool_


def _is_fancy_index(index_obj):
    return isinstance(index_obj, FANCY_INDEX_TYPES) and index_obj.dtype != np.bool_


def _is_create_view(index):
    # is view if all of index is slice, int or newaxis
    return all(isinstance(ind, (slice, Integral)) or ind is None for ind in index)


def _calc_order(a, index):
    if a.order == TensorOrder.C_ORDER:
        return TensorOrder.C_ORDER

    in_axis = 0
    for ind in index:
        if _is_bool_index(ind):
            in_axis += ind.ndim
            return TensorOrder.C_ORDER
        elif _is_fancy_index(ind):
            in_axis += 1
            return TensorOrder.C_ORDER
        elif ind is None:
            continue
        elif isinstance(ind, slice):
            shape = a.shape[in_axis]
            slc = ind.indices(shape)
            if slc[0] == 0 and slc[1] == shape and slc[2] == 1:
                continue
            else:
                return TensorOrder.C_ORDER
        else:
            assert isinstance(ind, Integral)
            in_axis += 1
            return TensorOrder.C_ORDER

    return TensorOrder.F_ORDER


def _getitem_nocheck(a, item, convert_bool_to_fancy=None):
    index = process_index(a.ndim, item, convert_bool_to_fancy=convert_bool_to_fancy)
    if convert_bool_to_fancy is False:
        # come from __setitem__, the bool index is not converted to fancy index
        # if multiple bool indexes or bool + fancy indexes exist,
        # thus the shape will be wrong,
        # here we just convert when calculating shape,
        # refer to issue #1282.
        shape = calc_shape(a.shape, process_index(a.ndim, index))
    else:
        shape = calc_shape(a.shape, index)
    tensor_order = _calc_order(a, index)
    op = TensorIndex(
        dtype=a.dtype,
        sparse=a.issparse(),
        indexes=list(index),
        create_view=_is_create_view(index),
    )
    return op(a, index, tuple(shape), order=tensor_order)


def _getitem(a, item):
    if isinstance(item, (list, tuple)) and all(
        isinstance(it, slice) and it == slice(None) for it in item
    ):
        # nothing to do
        return a

    # TODO(jisheng): field access, e.g. t['a'], t[['a', 'b']]
    return _getitem_nocheck(a, item)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from collections.abc import MutableMapping

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ...core import OutputType, recursive_tile
from ...core.custom_log import redirect_custom_log
from ...serialization.serializables import KeyField, AnyField, StringField
from ...utils import has_unknown_shape, enter_current_session, quiet_stdio
from ..core import SERIES_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import build_series


class DataFrameMap(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.MAP

    _input = KeyField("input")
    _arg = AnyField("arg")
    _na_action = StringField("na_action")

    def __init__(
        self, arg=None, na_action=None, output_types=None, memory_scale=None, **kw
    ):
        super().__init__(
            _arg=arg,
            _na_action=na_action,
            _output_types=output_types,
            _memory_scale=memory_scale,
            **kw
        )
        if not self.output_types:
            self.output_types = [OutputType.series]

    @property
    def input(self):
        return self._input

    @property
    def arg(self):
        return self._arg

    @property
    def na_action(self):
        return self._na_action

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._input = self._inputs[0]
        if len(inputs) == 2:
            self._arg = self._inputs[1]

    def __call__(self, series, dtype):
        if dtype is None:
            inferred_dtype = None
            if callable(self._arg):
                # arg is a function, try to inspect the signature
                sig = inspect.signature(self._arg)
                return_type = sig.return_annotation
                if return_type is not inspect._empty:
                    inferred_dtype = np.dtype(return_type)
                else:
                    try:
                        with quiet_stdio():
                            # try to infer dtype by calling the function
                            inferred_dtype = (
                                build_series(series)
                                .map(self._arg, na_action=self._na_action)
                                .dtype
                            )
                    except:  # noqa: E722  # nosec
                        pass
            else:
                if isinstance(self._arg, MutableMapping):
                    inferred_dtype = pd.Series(self._arg).dtype
                else:
                    inferred_dtype = self._arg.dtype
            if inferred_dtype is not None and np.issubdtype(inferred_dtype, np.number):
                if np.issubdtype(inferred_dtype, np.inexact):
                    # for the inexact e.g. float
                    # we can make the decision,
                    # but for int, due to the nan which may occur,
                    # we cannot infer the dtype
                    dtype = inferred_dtype
            else:
                dtype = inferred_dtype

        if dtype is None:
            raise ValueError(
                "cannot infer dtype, it needs to be specified manually for `map`"
            )
        else:
            dtype = np.int64 if dtype is int else dtype
            dtype = np.dtype(dtype)

        inputs = [series]
        if isinstance(self._arg, SERIES_TYPE):
            inputs.append(self._arg)

        if isinstance(series, SERIES_TYPE):
            return self.new_series(
                inputs,
                shape=series.shape,
                dtype=dtype,
                index_value=series.index_value,
                name=series.name,
            )
        else:
            return self.new_index(
                inputs,
                shape=series.shape,
                dtype=dtype,
                index_value=series.index_value,
                name=series.name,
            )

    @classmethod
    def tile(cls, op):
        in_series = op.input
        out_series = op.outputs[0]

        arg = op.arg
        if len(op.inputs) == 2:
            # make sure arg has known shape when it's a md.Series
            if has_unknown_shape(op.arg):
                yield
            arg = yield from recursive_tile(op.arg.rechunk(op.arg.shape))

        out_chunks = []
        for chunk in in_series.chunks:
            chunk_op = op.copy().reset_key()
            chunk_op.tileable_op_key = op.key
            chunk_inputs = [chunk]
            if len(op.inputs) == 2:
                chunk_inputs.append(arg.chunks[0])
            out_chunk = chunk_op.new_chunk(
                chunk_inputs,
                shape=chunk.shape,
                dtype=out_series.dtype,
                index_value=chunk.index_value,
                name=out_series.name,
                index=chunk.index,
            )
            out_chunks.append(out_chunk)

        new_op = op.copy()
        params = out_series.params
        params["chunks"] = out_chunks
        params["nsplits"] = in_series.nsplits
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    @redirect_custom_log
    @enter_current_session
    def execute(cls, ctx, op):
        series = ctx[op.inputs[0].key]
        out = op.outputs[0]
        if len(op.inputs) == 2:
            arg = ctx[op.inputs[1].key]
        else:
            arg = op.arg

        ret = series.map(arg, na_action=op.na_action)
        if ret.dtype != out.dtype:
            ret = ret.astype(out.dtype)
        ctx[out.key] = ret


def series_map(series, arg, na_action=None, dtype=None, memory_scale=None):
    """
    Map values of Series according to input correspondence.

    Used for substituting each value in a Series with another value,
    that may be derived from a function, a ``dict`` or
    a :class:`Series`.

    Parameters
    ----------
    arg : function, collections.abc.Mapping subclass or Series
        Mapping correspondence.
    na_action : {None, 'ignore'}, default None
        If 'ignore', propagate NaN values, without passing them to the
        mapping correspondence.
    dtype : np.dtype, default None
        Specify return type of the function. Must be specified when
        we cannot decide the return type of the function.
    memory_scale : float
        Specify the scale of memory uses in the function versus
        input size.

    Returns
    -------
    Series
        Same index as caller.

    See Also
    --------
    Series.apply : For applying more complex functions on a Series.
    DataFrame.apply : Apply a function row-/column-wise.
    DataFrame.applymap : Apply a function elementwise on a whole DataFrame.

    Notes
    -----
    When ``arg`` is a dictionary, values in Series that are not in the
    dictionary (as keys) are converted to ``NaN``. However, if the
    dictionary is a ``dict`` subclass that defines ``__missing__`` (i.e.
    provides a method for default values), then this default is used
    rather than ``NaN``.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> import mars.dataframe as md
    >>> s = md.Series(['cat', 'dog', mt.nan, 'rabbit'])
    >>> s.execute()
    0      cat
    1      dog
    2      NaN
    3   rabbit
    dtype: object

    ``map`` accepts a ``dict`` or a ``Series``. Values that are not found
    in the ``dict`` are converted to ``NaN``, unless the dict has a default
    value (e.g. ``defaultdict``):

    >>> s.map({'cat': 'kitten', 'dog': 'puppy'}).execute()
    0   kitten
    1    puppy
    2      NaN
    3      NaN
    dtype: object

    It also accepts a function:

    >>> s.map('I am a {}'.format).execute()
    0       I am a cat
    1       I am a dog
    2       I am a nan
    3    I am a rabbit
    dtype: object

    To avoid applying the function to missing values (and keep them as
    ``NaN``) ``na_action='ignore'`` can be used:

    >>> s.map('I am a {}'.format, na_action='ignore').execute()
    0     I am a cat
    1     I am a dog
    2            NaN
    3  I am a rabbit
    dtype: object
    """
    op = DataFrameMap(arg=arg, na_action=na_action, memory_scale=memory_scale)
    return op(series, dtype=dtype)


def index_map(idx, mapper, na_action=None, dtype=None, memory_scale=None):
    """
    Map values using input correspondence (a dict, Series, or function).

    Parameters
    ----------
    mapper : function, dict, or Series
        Mapping correspondence.
    na_action : {None, 'ignore'}
        If 'ignore', propagate NA values, without passing them to the
        mapping correspondence.
    dtype : np.dtype, default None
        Specify return type of the function. Must be specified when
        we cannot decide the return type of the function.
    memory_scale : float
        Specify the scale of memory uses in the function versus
        input size.

    Returns
    -------
    applied : Union[Index, MultiIndex], inferred
        The output of the mapping function applied to the index.
        If the function returns a tuple with more than one element
        a MultiIndex will be returned.
    """
    op = DataFrameMap(arg=mapper, na_action=na_action, memory_scale=memory_scale)
    return op(idx, dtype=dtype)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import AsyncGenerator

from .._utils cimport to_str, new_random_id
from .core cimport ActorRef, LocalActorRef


cpdef bytes new_actor_id():
    return new_random_id(32)


def create_actor_ref(*args, **kwargs):
    """
    Create an actor reference.

    Returns
    -------
    ActorRef
    """

    cdef str address
    cdef object uid
    cdef ActorRef existing_ref

    address = to_str(kwargs.pop('address', None))
    uid = kwargs.pop('uid', None)

    if kwargs:
        raise ValueError('Only `address` or `uid` keywords are supported')

    if len(args) == 2:
        if address:
            raise ValueError('address has been specified')
        address = to_str(args[0])
        uid = args[1]
    elif len(args) == 1:
        tp0 = type(args[0])
        if tp0 is ActorRef or tp0 is LocalActorRef:
            existing_ref = <ActorRef>(args[0])
            uid = existing_ref.uid
            address = to_str(address or existing_ref.address)
        else:
            uid = args[0]

    if uid is None:
        raise ValueError('Actor uid should be provided')

    return ActorRef(address, uid)


cdef set _is_async_generator_typecache = set()


cdef bint is_async_generator(obj):
    cdef type tp = type(obj)
    if tp in _is_async_generator_typecache:
        return True

    if isinstance(obj, AsyncGenerator):
        if len(_is_async_generator_typecache) < 100:
            _is_async_generator_typecache.add(tp)
        return True
    else:
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import AnyField
from ...config import options
from ..utils import decide_chunk_sizes
from .core import TensorNoInput
from ..array_utils import create_array


class TensorArange(TensorNoInput):
    _op_type_ = OperandDef.TENSOR_ARANGE

    _start = AnyField("start")
    _stop = AnyField("stop")
    _step = AnyField("step")

    def __init__(self, start=None, stop=None, step=None, dtype=None, **kw):
        if dtype is not None:
            dtype = np.dtype(dtype)
        elif stop is not None and step is not None:
            dtype = (
                np.dtype(dtype)
                if dtype is not None
                else np.arange(0, type(stop)(1), step).dtype
            )
        super().__init__(_start=start, _stop=stop, _step=step, dtype=dtype, **kw)

    @property
    def start(self):
        return self._start

    @property
    def stop(self):
        return self._stop

    @property
    def step(self):
        return self._step

    def to_chunk_op(self, *args):
        op = self.copy().reset_key()
        start, stop, step = args
        op._start = start
        op._stop = stop
        op._step = step
        return op

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]

        chunk_length = tensor.extra_params.raw_chunk_size or options.chunk_size
        chunk_length = decide_chunk_sizes(
            tensor.shape, chunk_length, tensor.dtype.itemsize
        )

        start, stop, step = op.start, op.stop, op.step  # noqa: F841

        out_chunks = []
        n_elem = 0
        for i, cs in enumerate(chunk_length[0]):
            chunk_start = start + n_elem * step
            chunk_stop = start + (n_elem + cs) * step
            chunk_size = max(int(np.ceil((chunk_stop - chunk_start) / step)), 0)
            if chunk_size > cs:
                chunk_stop -= step
            chunk_shape = (cs,)
            chunk_idx = (i,)
            chunk_op = op.to_chunk_op(chunk_start, chunk_stop, step)
            out_chunk = chunk_op.new_chunk(None, shape=chunk_shape, index=chunk_idx)
            n_elem += cs
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs,
            tensor.shape,
            order=tensor.order,
            chunks=out_chunks,
            nsplits=chunk_length,
        )

    @classmethod
    def execute(cls, ctx, op):
        ctx[op.outputs[0].key] = create_array(op)(
            "arange", op.start, op.stop, op.step, dtype=op.dtype
        )


def arange(*args, **kwargs):
    """
    Return evenly spaced values within a given interval.

    Values are generated within the half-open interval ``[start, stop)``
    (in other words, the interval including `start` but excluding `stop`).
    For integer arguments the function is equivalent to the Python built-in
    `range <http://docs.python.org/lib/built-in-funcs.html>`_ function,
    but returns a tensor rather than a list.

    When using a non-integer step, such as 0.1, the results will often not
    be consistent.  It is better to use ``linspace`` for these cases.

    Parameters
    ----------
    start : number, optional
        Start of interval.  The interval includes this value.  The default
        start value is 0.
    stop : number
        End of interval.  The interval does not include this value, except
        in some cases where `step` is not an integer and floating point
        round-off affects the length of `out`.
    step : number, optional
        Spacing between values.  For any output `out`, this is the distance
        between two adjacent values, ``out[i+1] - out[i]``.  The default
        step size is 1.  If `step` is specified as a position argument,
        `start` must also be given.
    dtype : dtype
        The type of the output tensor.  If `dtype` is not given, infer the data
        type from the other input arguments.
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default

    Returns
    -------
    arange : Tensor
        Tensor of evenly spaced values.

        For floating point arguments, the length of the result is
        ``ceil((stop - start)/step)``.  Because of floating point overflow,
        this rule may result in the last element of `out` being greater
        than `stop`.

    See Also
    --------
    linspace : Evenly spaced numbers with careful handling of endpoints.
    ogrid: Tensors of evenly spaced numbers in N-dimensions.
    mgrid: Grid-shaped tensors of evenly spaced numbers in N-dimensions.

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.arange(3).execute()
    array([0, 1, 2])
    >>> mt.arange(3.0).execute()
    array([ 0.,  1.,  2.])
    >>> mt.arange(3,7).execute()
    array([3, 4, 5, 6])
    >>> mt.arange(3,7,2).execute()
    array([3, 5])
    """
    kw_args = [kwargs.get("start"), kwargs.get("stop"), kwargs.get("step")]
    kw_def = any(arg is not None for arg in kw_args)
    dtype = None
    if not kw_def:
        if len(args) == 1:
            start = 0
            stop = args[0]
            step = 1
        elif len(args) == 2:
            start = args[0]
            stop = args[1]
            step = 1
        elif len(args) == 3:
            start, stop, step = args
        elif len(args) == 4:
            start, stop, step, dtype = args
            dtype = np.dtype(dtype)
        else:
            raise TypeError("Required argument 'start' (pos 1) not found")
    else:
        names = "start", "stop", "step"
        for i, arg in enumerate(args):
            if kw_args[i] is not None:
                raise TypeError(
                    f"Argument given by name ('{names[i]}') and position ({i})"
                )
            kw_args[i] = arg
        start, stop, step = kw_args

    if dtype is None:
        if "dtype" in kwargs:
            dtype = np.dtype(kwargs["dtype"])
        else:
            dtype = np.arange(0, type(stop)(1), step).dtype

    start, stop = dtype.type(start), dtype.type(stop)
    if dtype == np.datetime64 and not start:
        raise ValueError(
            "arange requires both a start and a stop for Mars datetime64 ranges"
        )
    if dtype == np.datetime64:
        span = np.array([stop - start])
        span[0] = step
        step = span[0]
        dtype = np.dtype(stop.dtype)
    else:
        step = dtype.type(step)
    size = max(int(np.ceil(np.true_divide(stop - start, step))), 0)

    op = TensorArange(start, stop, step, dtype=dtype, gpu=kwargs.get("gpu", False))
    shape = (size,)
    return op(shape, chunk_size=kwargs.pop("chunk_size", None))



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from typing import List, Tuple

import numpy as np

from ...core import OBJECT_TYPE, OBJECT_CHUNK_TYPE
from ...dataframe.core import (
    DtypesValue,
    IndexValue,
    DATAFRAME_TYPE,
    SERIES_TYPE,
    INDEX_TYPE,
    DATAFRAME_GROUPBY_TYPE,
    SERIES_GROUPBY_TYPE,
    DATAFRAME_CHUNK_TYPE,
    SERIES_CHUNK_TYPE,
    INDEX_CHUNK_TYPE,
    DATAFRAME_GROUPBY_CHUNK_TYPE,
    SERIES_GROUPBY_CHUNK_TYPE,
    CATEGORICAL_TYPE,
    CATEGORICAL_CHUNK_TYPE,
)
from ...tensor.core import TensorOrder, TENSOR_TYPE, TENSOR_CHUNK_TYPE
from ...utils import dataslots
from .core import (
    _TileableMeta,
    _ChunkMeta,
    PandasDtypeType,
    register_meta_type,
)


"""
Create a separate module for metas to avoid direct
dependency on mars.dataframe
"""


@register_meta_type(TENSOR_TYPE)
@dataslots
@dataclass
class TensorMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: np.dtype = None
    order: TensorOrder = None


@register_meta_type(DATAFRAME_TYPE)
@dataslots
@dataclass
class DataFrameMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None


@register_meta_type(SERIES_TYPE)
@dataslots
@dataclass
class SeriesMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@register_meta_type(INDEX_TYPE)
@dataslots
@dataclass
class IndexMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@register_meta_type(DATAFRAME_GROUPBY_TYPE)
@dataslots
@dataclass
class DataFrameGroupByMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None
    selection: List = None


@register_meta_type(SERIES_GROUPBY_TYPE)
@dataslots
@dataclass
class SeriesGroupByMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None
    selection: List = None


@register_meta_type(CATEGORICAL_TYPE)
@dataslots
@dataclass
class CategoricalMeta(_TileableMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    categories_value: IndexValue = None


@register_meta_type(OBJECT_TYPE)
@dataslots
@dataclass
class ObjectMeta(_TileableMeta):
    pass


@register_meta_type(TENSOR_CHUNK_TYPE)
@dataslots
@dataclass
class TensorChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: np.dtype = None
    order: TensorOrder = None


@register_meta_type(DATAFRAME_CHUNK_TYPE)
@dataslots
@dataclass
class DataFrameChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None


@register_meta_type(SERIES_CHUNK_TYPE)
@dataslots
@dataclass
class SeriesChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@register_meta_type(INDEX_CHUNK_TYPE)
@dataslots
@dataclass
class IndexChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None


@register_meta_type(DATAFRAME_GROUPBY_CHUNK_TYPE)
@dataslots
@dataclass
class DataFrameGroupByChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtypes_value: DtypesValue = None
    index_value: IndexValue = None
    selection: List = None


@register_meta_type(SERIES_GROUPBY_CHUNK_TYPE)
@dataslots
@dataclass
class SeriesGroupByChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    index_value: IndexValue = None
    selection: List = None


@register_meta_type(CATEGORICAL_CHUNK_TYPE)
@dataslots
@dataclass
class CategoricalChunkMeta(_ChunkMeta):
    shape: Tuple[int] = None
    dtype: PandasDtypeType = None
    categories_value: IndexValue = None


@register_meta_type(OBJECT_CHUNK_TYPE)
@dataslots
@dataclass
class ObjectChunkMeta(_ChunkMeta):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from ... import opcodes as OperandDef
from ..datasource import tensor as astensor
from .core import TensorReduction, TensorReductionMixin


class TensorNanSum(TensorReduction, TensorReductionMixin):
    _op_type_ = OperandDef.NANSUM
    _func_name = "nansum"

    def __init__(self, axis=None, keepdims=None, combine_size=None, stage=None, **kw):
        stage = self._rewrite_stage(stage)
        super().__init__(
            _axis=axis,
            _keepdims=keepdims,
            _combine_size=combine_size,
            stage=stage,
            **kw
        )


def nansum(a, axis=None, dtype=None, out=None, keepdims=None, combine_size=None):
    """
    Return the sum of array elements over a given axis treating Not a
    Numbers (NaNs) as zero.

    Zero is returned for slices that are all-NaN or
    empty.

    Parameters
    ----------
    a : array_like
        Tensor containing numbers whose sum is desired. If `a` is not an
        tensor, a conversion is attempted.
    axis : int, optional
        Axis along which the sum is computed. The default is to compute the
        sum of the flattened array.
    dtype : data-type, optional
        The type of the returned tensor and of the accumulator in which the
        elements are summed.  By default, the dtype of `a` is used.  An
        exception is when `a` has an integer type with less precision than
        the platform (u)intp. In that case, the default will be either
        (u)int32 or (u)int64 depending on whether the platform is 32 or 64
        bits. For inexact inputs, dtype must be inexact.
    out : Tensor, optional
        Alternate output tensor in which to place the result.  The default
        is ``None``. If provided, it must have the same shape as the
        expected output, but the type will be cast if necessary.  See
        `doc.ufuncs` for details. The casting of NaN to integer can yield
        unexpected results.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the original `a`.


        If the value is anything but the default, then
        `keepdims` will be passed through to the `mean` or `sum` methods
        of sub-classes of `Tensor`.  If the sub-classes methods
        does not implement `keepdims` any exceptions will be raised.
    combine_size: int, optional
        The number of chunks to combine.

    Returns
    -------
    nansum : Tensor.
        A new tensor holding the result is returned unless `out` is
        specified, in which it is returned. The result has the same
        size as `a`, and the same shape as `a` if `axis` is not None
        or `a` is a 1-d array.

    See Also
    --------
    mt.sum : Sum across tensor propagating NaNs.
    isnan : Show which elements are NaN.
    isfinite: Show which elements are not NaN or +/-inf.

    Notes
    -----
    If both positive and negative infinity are present, the sum will be Not
    A Number (NaN).

    Examples
    --------
    >>> import mars.tensor as mt

    >>> mt.nansum(1).execute()
    1
    >>> mt.nansum([1]).execute()
    1
    >>> mt.nansum([1, mt.nan]).execute()
    1.0
    >>> a = mt.array([[1, 1], [1, mt.nan]])
    >>> mt.nansum(a).execute()
    3.0
    >>> mt.nansum(a, axis=0).execute()
    array([ 2.,  1.])
    >>> mt.nansum([1, mt.nan, mt.inf]).execute()
    inf
    >>> mt.nansum([1, mt.nan, mt.NINF]).execute()
    -inf
    >>> mt.nansum([1, mt.nan, mt.inf, -mt.inf]).execute() # both +/- infinity present
    nan

    """
    a = astensor(a)
    if dtype is None:
        dtype = np.nansum(np.empty((1,), dtype=a.dtype)).dtype
    op = TensorNanSum(
        axis=axis, dtype=dtype, keepdims=keepdims, combine_size=combine_size
    )
    return op(a, out=out)
